# Copyright (c) Microsoft. All rights reserved.

import asyncio
import logging
import multiprocessing
import signal
import time
import warnings
from typing import Any, List, Optional, TypeVar, Union

from mantisdk.adapter import TraceAdapter, TracerTraceToTriplet
from mantisdk.algorithm import Algorithm
from mantisdk.client import MantisdkClient
from mantisdk.litagent import LitAgent
from mantisdk.runner import LegacyAgentRunner
from mantisdk.tracer.base import Tracer
from mantisdk.types import Dataset, ParallelWorkerBase

logger = logging.getLogger(__name__)

T_co = TypeVar("T_co", covariant=True)


class TrainerLegacy(ParallelWorkerBase):
    """Trainer for legacy mode for v0.1 compatibility."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize the TrainerLegacy.

        This method is mainly to make type checker happy.
        It won't be used in practice.
        """
        self._dev = kwargs.pop("dev", False)
        self.algorithm: Optional[Algorithm] = kwargs.pop("algorithm", None)
        self.tracer: Tracer = kwargs.pop("tracer", None)
        self.n_workers: int = kwargs.pop("n_workers", None)
        self.max_tasks: Optional[int] = kwargs.pop("max_tasks", None)
        self.daemon: bool = kwargs.pop("daemon", True)
        self.triplet_exporter: TraceAdapter[Any] = kwargs.pop("triplet_exporter", None)

    def _extract_client_from_data(
        self, data: Union[str, MantisdkClient, Dataset[Any]]
    ) -> Optional[MantisdkClient]:
        """Extract client from data if it's a string URL or MantisdkClient."""
        if isinstance(data, str):
            if not data.startswith("http://") and not data.startswith("https://"):
                raise ValueError("String data must be a valid URL starting with http:// or https://")
            return MantisdkClient(endpoint=data)
        elif isinstance(data, MantisdkClient):
            return data
        return None

    def _extract_dataset_from_data(
        self, data: Union[str, MantisdkClient, Dataset[Any]]
    ) -> Optional[Dataset[Any]]:
        """Extract dataset from data if it's a Dataset."""
        if isinstance(data, str) or isinstance(data, MantisdkClient):
            return None
        return data

    def _determine_backend(
        self,
        train_data: Union[str, MantisdkClient, Dataset[Any]],
        dev_data: Union[str, MantisdkClient, Dataset[Any], None] = None,
    ) -> Union[str, MantisdkClient]:
        """Determine which backend to use for initialization."""
        if self._dev:
            if dev_data is None:
                raise ValueError("dev_data must be provided when dev=True.")
            client = self._extract_client_from_data(dev_data)
            if client is None:
                raise ValueError("dev_data must be a string URL or MantisdkClient when dev=True.")
            return client
        else:
            client = self._extract_client_from_data(train_data)
            if client is None and self.algorithm is None:
                raise ValueError(
                    "train_data must be a string URL or MantisdkClient when no algorithm is provided."
                )
            elif client is None and self.algorithm is not None:
                # Algorithm will be responsible for creating the client
                client = self.algorithm.get_client()
                logger.info(f"Algorithm created client: {client}")
                return client
            if client is None:
                raise ValueError(
                    "train_data must be a string URL or MantisdkClient when no algorithm is provided."
                )
            return client

    def init(self, backend: Union[str, MantisdkClient]) -> None:
        logger.info(f"Initializing Trainer...")

        self._init_client(backend)

        self.tracer.init()

        logger.info(f"Trainer main initialization complete.")

    def teardown(self) -> None:
        logger.info(f"Cleaning up Trainer...")
        self.tracer.teardown()

        self._client = None
        logger.info(f"Trainer main cleanup complete.")

    def client(self) -> MantisdkClient:
        """Returns the MantisdkClient instance."""
        if self._client is None:
            raise RuntimeError("MantisdkClient has not been initialized. Call `init` first.")
        return self._client

    def _init_client(self, backend: Union[str, MantisdkClient]) -> MantisdkClient:
        if self._client is None:
            if isinstance(backend, MantisdkClient):
                logger.info("Using provided MantisdkClient instance.")
                self._client = backend
            else:
                logger.info(f"Initializing MantisdkClient with endpoint: {backend}")
                if not isinstance(backend, str):  # type: ignore
                    raise ValueError("backend must be a string URL or an MantisdkClient instance.")
                if not backend.startswith("http://") and not backend.startswith("https://"):
                    raise ValueError("backend must be a valid URL starting with http:// or https://")
                # Initialize the client with the provided backend URL
                self._client = MantisdkClient(endpoint=backend)
        else:
            logger.warning("MantisdkClient already initialized. Returning existing instance.")
        return self._client

    def _worker_main_loop(self, agent: LitAgent[Any], worker_id: int, is_async: bool):
        """The main function for each worker process.

        This function initializes the client and the loop, then starts the
        execution. It also configures process-specific settings like the
        process title and signal handling.

        Args:
            agent: The `LitAgent` instance to run.
            worker_id: The unique ID for this worker.
            is_async: A boolean indicating if the async loop should be run.
        """
        if self.n_workers > 1:
            import setproctitle

            # Ignore Ctrl+C in worker processes; the main process handles it
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            setproctitle.setproctitle(multiprocessing.current_process().name)

        # Now we are in child processes, so we can safely set up the environment.
        agent.set_trainer(self)  # type: ignore
        if not isinstance(self.triplet_exporter, TracerTraceToTriplet):  # type: ignore
            raise ValueError("triplet_exporter must be a TracerTraceToTriplet for the legacy trainer.")
        # TODO: this should be set elsewhere
        if agent.trained_agents:
            self.triplet_exporter.agent_match = agent.trained_agents
        self._initialize_worker_env(worker_id)

        mode = "Async" if is_async else "Sync"
        logger.info(f"[Worker {worker_id}] {mode} worker process started.")

        num_processed = 0

        try:
            client = self.client()
            loop = LegacyAgentRunner(
                agent=agent,
                client=client,
                tracer=self.tracer,
                triplet_exporter=self.triplet_exporter,
                max_tasks=self.max_tasks,
                worker_id=worker_id,
            )
            loop.init_worker(worker_id)  # type: ignore
            if is_async:
                num_processed = asyncio.run(loop.iter_async())
            else:
                num_processed = loop.iter()
        except Exception:
            logger.exception(f"[Worker {worker_id}] Unhandled exception in worker loop.")
        finally:
            self._teardown_worker_env(worker_id)

        return num_processed

    def _initialize_worker_env(self, worker_id: int):
        logger.info(f"[Worker {worker_id}] Setting up trainer environment...")  # worker_id included in process name
        self.tracer.init_worker(worker_id)

    def _teardown_worker_env(self, worker_id: int):
        logger.info(f"[Worker {worker_id}] Cleaning up trainer environment...")
        self.tracer.teardown_worker(worker_id)
        logger.info(f"[Worker {worker_id}] Environment cleanup complete.")

    @staticmethod
    def kill_orphaned_processes() -> None:
        """
        Kill any orphaned processes that may have been left behind by previous runs.
        This is useful for cleaning up after crashes or unexpected exits.
        """
        import psutil

        for proc in psutil.process_iter():  # type: ignore
            # check whether the process name matches
            if proc.name().startswith("Mantisdk-"):
                proc.kill()

    def _terminate_processes(self, processes: List[multiprocessing.Process]) -> None:
        if self.n_workers > 1 and len(processes) > 0:
            for i, p in enumerate(processes):
                if p.is_alive():
                    logger.info(f"Terminating worker {i} (name: {p.name}, PID: {p.pid})...")
                    p.terminate()
                else:
                    logger.info(f"Worker {i} (name: {p.name}, PID: {p.pid}) is not alive or has already terminated.")
            for i, p in enumerate(processes):
                if p.is_alive():
                    p.join(timeout=10)  # Give some time to terminate
                if p.is_alive():  # If still alive, kill
                    logger.warning(
                        f"Worker {i} (name: {p.name}, PID: {p.pid}) did not terminate gracefully, killing..."
                    )
                    p.kill()
                    p.join(timeout=10)  # Ensure it's reaped

    def fit_v0(
        self,
        agent: LitAgent[T_co],
        train_data: Union[str, MantisdkClient, Dataset[T_co]],
        *,
        val_data: Union[str, MantisdkClient, Dataset[T_co], None] = None,
        dev_data: Union[str, MantisdkClient, Dataset[T_co], None] = None,
        dev_backend: Union[str, MantisdkClient, None] = None,
    ):
        """Train the agent using the provided data.

        Each data argument can be a string URL connecting to a mantisdk server,
        or an MantisdkClient instance connecting to a server (or mock server), or a dataset.
        If no algorithm is provided when instantiating the trainer, the data must be
        provided to connecting a server. Otherwise, dataset is also allowed and will be
        passed to the algorithm.

        If the algorithm is instantiated and there is no URL/client provided,
        the algorithm will be responsible for creating a client that will connect to itself.
        It can also create a mock client if the algorithm does not require a server.
        """

        if dev_backend is not None:
            warnings.warn("dev_backend is deprecated. Use dev_data instead.")
            if dev_data is not None:
                raise ValueError("dev_data and dev_backend cannot be provided at the same time.")
            dev_data = dev_backend

        # Extract datasets for algorithm if available
        train_dataset = self._extract_dataset_from_data(train_data)
        val_dataset = self._extract_dataset_from_data(val_data) if val_data else None

        # Initialize the algorithm with trainer if provided
        if self.algorithm is not None:
            self.algorithm.set_trainer(self)  # type: ignore
            # DO NOT RUN TRAINING HERE. Need to spawn the worker first.

        # Determine the backend to use for client-server mode
        backend = self._determine_backend(train_data, dev_data)

        if self._dev:
            logger.warning(f"Running in dev mode. Using dev backend: {backend}")
        else:
            logger.debug(f"Running in non-dev mode. Using backend: {backend}")

        self.init(backend)

        processes: List[multiprocessing.Process] = []

        # Determine if the agent is asynchronous

        mode = "asynchronous" if agent.is_async() else "synchronous"

        try:
            if self.n_workers == 1:
                logger.info(f"Running with n_workers=1 ({mode} in main process).")

                # Warn if algorithm is set with single worker mode
                if self.algorithm is not None:
                    logger.warning(
                        "Algorithm is set but using single worker mode. Algorithm will never get the chance to run."
                    )
                    # Ideally the single worker should be run in a separate thread or process.

                num_tasks = self._worker_main_loop(agent, 0, agent.is_async())
                logger.info(f"Single worker mode finished. Tasks processed: {num_tasks}")

                # If algorithm is provided and we have datasets, run algorithm after worker completes
                if self.algorithm is not None and train_dataset is not None:
                    logger.info("Running algorithm training after worker completion.")
                    self.algorithm.run(
                        train_dataset=train_dataset,
                        val_dataset=val_dataset,
                    )
            else:
                logger.info(f"Running with n_workers={self.n_workers} ({mode} multiprocessing).")
                for i in range(self.n_workers):
                    process_name = f"Mantisdk-Worker-{i}"
                    p = multiprocessing.Process(
                        target=self._worker_main_loop,
                        args=(agent, i, agent.is_async()),
                        daemon=self.daemon,
                        name=process_name,
                    )
                    processes.append(p)
                    logger.info(f"Starting worker process {i} (name: {process_name})...")
                    p.start()

                if self.daemon:
                    # If algorithm is provided and we have datasets, pass them to the algorithm
                    if self.algorithm is not None:
                        logger.info("All workers have been spawned. Running algorithm training with provided datasets.")
                        self.algorithm.run(
                            train_dataset=train_dataset,
                            val_dataset=val_dataset,
                        )
                        logger.info("Algorithm exits. Killing the workers.")
                        self._terminate_processes(processes)

                    for i, p in enumerate(processes):
                        p.join()  # Wait for the process to complete
                        logger.info(
                            f"Worker process {i} (name: {p.name}, PID: {p.pid}) joined with exit code {p.exitcode}."
                        )
                        if p.exitcode != 0:
                            logger.warning(
                                f"Worker process {i} (name: {p.name}, PID: {p.pid}) exited with non-zero code: {p.exitcode}."
                            )

                    logger.info(f"All {self.n_workers} worker processes have completed.")
                else:
                    logger.info("All worker processes started. Main process will not wait.")

                    # A hack to stop the main process from waiting for child processes to finish.
                    time.sleep(1)  # Give workers time to start
                    import multiprocessing.process as multiprocessing_process

                    multiprocessing_process._children.clear()  # type: ignore

                    if self.algorithm is not None:
                        logger.info("Main process continues to run algorithm.")
                        self.algorithm.run(
                            train_dataset=train_dataset,
                            val_dataset=val_dataset,
                        )
                        logger.info("Algorithm exits. Killing the workers.")
                        self._terminate_processes(processes)

        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt received. Killing the workers.")
            self._terminate_processes(processes)
            logger.info(f"Workers terminated or single worker interrupted.")
            raise
        except Exception:
            logger.exception(f"Unhandled exception in fit method.")
            self._terminate_processes(processes)
            logger.info(f"Workers terminated or single worker interrupted.")
            raise
        finally:
            if self.daemon:
                self.teardown()
            else:
                logger.info("Main process exiting. Please use Trainer.kill_orphaned_processes() for cleanup.")



# Copyright (c) Microsoft. All rights reserved.

import hashlib
import uuid

__all__ = ["generate_id"]


def generate_id(length: int) -> str:
    """Generate a random ID of the given length.

    Args:
        length: The length of the ID to generate.

    Returns:
        A random ID of the given length.
    """
    return hashlib.sha1(uuid.uuid4().bytes).hexdigest()[:length]



# Copyright (c) Microsoft. All rights reserved.

"""Tracing configuration for algorithms.

This module provides a standardized way for algorithms to define their
tracing metadata (environment and tags) that flows through to Mantis/Insight.

It also provides infrastructure for call-type tagging, allowing algorithms
to define decorators that tag LLM calls (e.g., @gepa.judge, @gepa.agent).
"""

import asyncio
import uuid
from contextvars import ContextVar
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Callable, List, Optional, TypeVar

# Type variable for decorated functions
F = TypeVar("F", bound=Callable[..., Any])

# ============================================================================
# Call Type Context Propagation
# ============================================================================

# Context variable for the current LLM call type (agent, judge, reflection, etc.)
# This is used by instrumentation to tag spans appropriately.
_call_type_context: ContextVar[Optional[str]] = ContextVar("mantis_call_type", default=None)


def get_current_call_type() -> Optional[str]:
    """Get the current LLM call type from context.
    
    Returns:
        The current call type (e.g., "agent-call", "judge-call") or None if not set.
    """
    return _call_type_context.get()


def inject_call_type_header(extra_headers: Optional[dict] = None) -> dict:
    """Inject x-mantis-call-type header from context into extra_headers.
    
    This reads the current call type from context (set by decorators like
    @gepa.agent, @gepa.judge) and adds it to extra_headers for LLM calls.
    The LLM proxy receives this header and tags the trace accordingly.
    
    Args:
        extra_headers: Optional existing headers dict to merge into.
    
    Returns:
        A dict with x-mantis-call-type header if call type is set.
    
    Example:
        >>> @gepa.agent
        >>> def my_agent(client):
        ...     return client.chat.completions.create(
        ...         model="gpt-4",
        ...         messages=[...],
        ...         extra_headers=inject_call_type_header()
        ...     )
    """
    if extra_headers is None:
        extra_headers = {}
    
    call_type = get_current_call_type()
    if call_type:
        extra_headers["x-mantis-call-type"] = call_type
    
    return extra_headers


def call_type_decorator(call_type: str) -> Callable[[F], F]:
    """Factory to create call-type tagging decorators.
    
    This is used by algorithms to define their own call-type decorators.
    When a decorated function executes, any LLM calls made within it
    will be tagged with the specified call type.
    
    Args:
        call_type: The call type tag (e.g., "agent-call", "judge-call").
    
    Returns:
        A decorator that sets the call type context for the decorated function.
    
    Example:
        >>> # In algorithm/gepa/__init__.py
        >>> agent = call_type_decorator("agent-call")
        >>> judge = call_type_decorator("judge-call")
        >>> 
        >>> # In user code
        >>> @gepa.judge
        >>> def grade_response(client, response, expected):
        ...     return client.chat.completions.parse(...)  # Tagged as "judge-call"
    """
    def decorator(fn: F) -> F:
        if asyncio.iscoroutinefunction(fn):
            @wraps(fn)
            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                token = _call_type_context.set(call_type)
                try:
                    return await fn(*args, **kwargs)
                finally:
                    _call_type_context.reset(token)
            return async_wrapper  # type: ignore
        else:
            @wraps(fn)
            def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
                token = _call_type_context.set(call_type)
                try:
                    return fn(*args, **kwargs)
                finally:
                    _call_type_context.reset(token)
            return sync_wrapper  # type: ignore
    return decorator


@dataclass
class TracingConfig:
    """Configuration for algorithm-specific tracing metadata.
    
    Each algorithm (GEPA, VERL, etc.) should define its own TracingConfig
    to identify its traces in Mantis. The environment, tags, and session_id are
    propagated through the LLM proxy and stored with traces.
    
    Example:
        >>> config = TracingConfig(
        ...     environment="mantisdk-gepa",
        ...     algorithm_name="gepa"
        ... )
        >>> config.get_tags("train")
        ['gepa', 'train']
    
    Attributes:
        environment: The environment name for traces (e.g., "mantisdk-gepa").
            This appears in Mantis's environment column.
        algorithm_name: The algorithm identifier used as the base tag.
        session_id: Unique session identifier for grouping related traces.
            Auto-generated if not provided.
    """
    
    environment: str
    algorithm_name: str
    session_id: str = field(default_factory=lambda: f"session-{uuid.uuid4().hex[:12]}")
    
    def get_tags(self, trace_type: str) -> List[str]:
        """Generate tags for a specific trace type.
        
        Args:
            trace_type: The type of trace (e.g., "train", "validation", 
                "reflection", "test").
        
        Returns:
            A list of tags: [algorithm_name, trace_type]
        """
        return [self.algorithm_name, trace_type]
    
    def get_detailed_tags(
        self,
        phase: str,
        extra_tags: Optional[List[str]] = None,
    ) -> List[str]:
        """Generate tags with execution context.
        
        Args:
            phase: Execution phase (e.g., "train", "validation", "reflection").
            extra_tags: Optional algorithm-specific tags to append.
        
        Returns:
            A list of tags, e.g.: ["gepa", "train", "custom-tag"]
        """
        tags = [self.algorithm_name, phase]
        
        # Add any algorithm-specific extra tags
        if extra_tags:
            tags.extend(extra_tags)
        
        return tags
    
    def to_metadata(self, trace_type: Optional[str] = None) -> dict:
        """Convert to a metadata dict for rollouts or LLM calls.
        
        Args:
            trace_type: Optional trace type to include in tags.
                If None, only the algorithm_name is included in tags.
        
        Returns:
            Dict with "environment", "tags", and "session_id" keys.
        """
        if trace_type:
            tags = self.get_tags(trace_type)
        else:
            tags = [self.algorithm_name]
        
        return {
            "environment": self.environment,
            "tags": tags,
            "session_id": self.session_id,
        }
    
    def to_detailed_metadata(
        self,
        phase: str,
        extra_tags: Optional[List[str]] = None,
    ) -> dict:
        """Convert to a metadata dict with execution context.
        
        Args:
            phase: Execution phase (e.g., "train", "validation", "reflection").
            extra_tags: Optional algorithm-specific tags to append.
        
        Returns:
            Dict with "environment", "tags", and "session_id" keys.
        """
        return {
            "environment": self.environment,
            "tags": self.get_detailed_tags(phase, extra_tags),
            "session_id": self.session_id,
        }



# Copyright (c) Microsoft. All rights reserved.

"""Convenient helpers for creating spans / traces.

All emitters operate in two modes, switchable via the `propagate` parameter.
The emitters first [`SpanCreationRequest`][mantisdk.SpanCreationRequest] object, then:

1. When `propagate` is True, this creation request will be propagated to the active tracer
   and a [`Span`][mantisdk.Span] instance will be created (possibly deferred).
2. When `propagate` is False, the creation request will be returned directly. Useful for cases
   when you don't have a tracer but you want to create a creation request for later use.
"""

from .annotation import emit_annotation, operation
from .exception import emit_exception
from .message import emit_message, get_message_value
from .object import emit_object, get_object_value
from .reward import (
    emit_reward,
    find_final_reward,
    find_reward_spans,
    get_reward_value,
    get_rewards_from_span,
    is_reward_span,
    reward,
)

__all__ = [
    "reward",
    "operation",
    "emit_reward",
    "get_reward_value",
    "get_rewards_from_span",
    "is_reward_span",
    "find_reward_spans",
    "find_final_reward",
    "emit_message",
    "emit_object",
    "emit_exception",
    "emit_annotation",
    "get_message_value",
    "get_object_value",
]



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import uuid
from typing import Any, Dict, List, Mapping, Optional, Sequence, TypedDict

from mantisdk.algorithm.gepa.lib.core.adapter import EvaluationBatch, GEPAAdapter

from mantisdk.adapter import TraceAdapter
from mantisdk.reward import find_final_reward, get_rewards_from_span, find_reward_spans
from mantisdk.store.base import LightningStore
from mantisdk.types import NamedResources, PromptTemplate, Rollout, Span, TracingConfig

from .tracing import GEPATracingContext

logger = logging.getLogger(__name__)


class MantisdkDataInst(TypedDict):
    """Data instance for Mantisdk."""

    input: Dict[str, Any]
    id: Optional[str]


class MantisdkTrajectory(TypedDict):
    """Trajectory for Mantisdk.
    
    Captures the data needed for reflection:
    - original_input: The task input data
    - assistant_response: The final LLM response
    - feedback: Evaluation feedback including score and status
    - spans: Raw spans for detailed analysis
    """

    rollout_id: str
    original_input: Dict[str, Any]  # The task input
    assistant_response: str  # The final LLM response
    feedback: str  # Feedback including score and error info
    spans: List[Span]


class MantisdkRolloutOutput(TypedDict):
    """Rollout output for Mantisdk."""

    final_reward: Optional[float]
    status: str


# TypedDict for GEPA's expected reflective record format
MantisdkReflectiveRecord = TypedDict(
    "MantisdkReflectiveRecord",
    {
        "Inputs": str,
        "Generated Outputs": str,
        "Feedback": str,
    },
)


class MantisdkGEPAAdapter(
    GEPAAdapter[MantisdkDataInst, MantisdkTrajectory, MantisdkRolloutOutput]
):
    """Adapter to bridge GEPA with Mantisdk.
    
    This adapter:
    1. Evaluates candidates by creating rollouts in the LightningStore
    2. Waits for runners to execute the agent
    3. Collects spans and rewards
    4. Builds reflective datasets for GEPA's reflection mechanism
    """

    def __init__(
        self,
        store: LightningStore,
        loop: asyncio.AbstractEventLoop,
        resource_name: str,
        adapter: TraceAdapter,
        llm_proxy_resource: Any = None,
        rollout_batch_timeout: float = 600.0,
        tracing_config: Optional[TracingConfig] = None,
        tracing_context: Optional[GEPATracingContext] = None,
    ) -> None:
        """Initialize the adapter.

        Args:
            store: The LightningStore instance.
            loop: The asyncio loop where the store and other async components are running.
            resource_name: The name of the resource to update (e.g., "prompt_template").
            adapter: The TraceAdapter to convert spans to messages (optional, for debugging).
            llm_proxy_resource: LLM resource from the proxy to include in resources.
            rollout_batch_timeout: Timeout for waiting for rollouts in seconds.
            tracing_config: Optional tracing configuration from the algorithm.
            tracing_context: Optional GEPA tracing context for detailed execution tracking.
        """
        self.store = store
        self.loop = loop
        self.resource_name = resource_name
        self.adapter = adapter
        self.llm_proxy_resource = llm_proxy_resource
        self.rollout_batch_timeout = rollout_batch_timeout
        self.tracing_config = tracing_config
        self.tracing_context = tracing_context

    def evaluate(
        self,
        batch: List[MantisdkDataInst],
        candidate: Dict[str, str],
        capture_traces: bool = False,
    ) -> EvaluationBatch[MantisdkTrajectory, MantisdkRolloutOutput]:
        """Evaluate a candidate on a batch of data instances.

        This method bridges the synchronous GEPA call to the asynchronous Mantisdk execution.
        """
        # Run the async evaluation in the main loop and wait for result
        future = asyncio.run_coroutine_threadsafe(
            self._evaluate_async(batch, candidate, capture_traces), self.loop
        )
        return future.result(timeout=self.rollout_batch_timeout + 60)  # Add buffer to timeout

    async def _evaluate_async(
        self,
        batch: List[MantisdkDataInst],
        candidate: Dict[str, str],
        capture_traces: bool = False,
    ) -> EvaluationBatch[MantisdkTrajectory, MantisdkRolloutOutput]:
        """Asynchronous evaluation logic."""
        # 1. Build resources from candidate
        resources: NamedResources = {}
        for key, value in candidate.items():
            resources[key] = PromptTemplate(template=value, engine="f-string")
        
        # Add the LLM resource if available
        if self.llm_proxy_resource is not None:
            resources["llm"] = self.llm_proxy_resource

        # Create a unique resource version for this evaluation
        candidate_version = f"gepa-{uuid.uuid4().hex[:8]}"
        await self.store.update_resources(candidate_version, resources)

        # 2. Enqueue rollouts for each item in the batch
        rollout_ids = []
        batch_inputs: Dict[str, Dict[str, Any]] = {}  # Map rollout_id -> original input
        
        # Prepare metadata from tracing config with detailed context
        metadata = None
        if self.tracing_config and self.tracing_context:
            # Generate candidate hash for tracking
            candidate_hash = hashlib.md5(str(candidate).encode()).hexdigest()[:8]
            self.tracing_context.set_candidate(candidate_hash)
            
            # Get batch item IDs for validation detection
            item_ids = [item.get("id", str(i)) for i, item in enumerate(batch)]
            
            # Detect if this is a validation batch
            if self.tracing_context.is_validation_batch(item_ids):
                self.tracing_context.set_phase("validation-eval")
            else:
                # Register training items on first batch of generation
                if self.tracing_context.batch_count == 0:
                    self.tracing_context.register_training_items(item_ids)
                self.tracing_context.set_phase("train-eval")
            
            # Get batch ID and build GEPA-specific tags
            batch_id = self.tracing_context.next_batch()
            gepa_tags = [
                f"gen-{self.tracing_context.generation}",
                f"candidate-{candidate_hash}",
                batch_id,
            ]
            metadata = self.tracing_config.to_detailed_metadata(
                phase=self.tracing_context.phase,
                extra_tags=gepa_tags,
            )
            logger.debug(
                f"Batch evaluation: session={self.tracing_context.session_id}, "
                f"phase={self.tracing_context.phase}, "
                f"gen={self.tracing_context.generation}, candidate={candidate_hash}, {batch_id}"
            )
        elif self.tracing_config:
            # Fallback to simple metadata if no context
            metadata = self.tracing_config.to_metadata("train")

        for item in batch:
            task_input = item["input"]
            res = await self.store.enqueue_rollout(
                input=task_input,
                mode="train",
                resources_id=candidate_version,
                metadata=metadata,
            )
            rollout_ids.append(res.rollout_id)
            batch_inputs[res.rollout_id] = task_input

        # 3. Wait for completion
        try:
            completed_rollouts = await self.store.wait_for_rollouts(
                rollout_ids=rollout_ids, timeout=self.rollout_batch_timeout
            )
        except Exception as e:
            logger.error(f"Error waiting for rollouts: {e}")
            # Return failure results for all items
            outputs: List[MantisdkRolloutOutput] = []
            trajectories: List[MantisdkTrajectory] = [] if capture_traces else None
            scores: List[float] = []
            
            for rollout_id in rollout_ids:
                outputs.append({"final_reward": 0.0, "status": "failed"})
                scores.append(0.0)
                if capture_traces:
                    trajectories.append({
                        "rollout_id": rollout_id,
                        "original_input": batch_inputs.get(rollout_id, {}),
                        "assistant_response": "",
                        "feedback": f"Rollout failed with error: {e}",
                        "spans": [],
                    })
            
            return EvaluationBatch(
                outputs=outputs,
                trajectories=trajectories,
                scores=scores,
            )

        # 4. Collect results
        outputs: List[MantisdkRolloutOutput] = []
        trajectories: List[MantisdkTrajectory] = [] if capture_traces else None
        scores: List[float] = []

        for rollout in completed_rollouts:
            # Query spans for this rollout
            spans = await self.store.query_spans(rollout.rollout_id)
            
            # Find final reward
            final_reward_val = find_final_reward(spans)
            if final_reward_val is None:
                final_reward_val = 0.0
            
            outputs.append({
                "final_reward": final_reward_val,
                "status": rollout.status,
            })
            scores.append(final_reward_val)
            
            if capture_traces:
                # Extract assistant response from spans
                assistant_response = self._extract_assistant_response(spans)
                
                # Build feedback string
                feedback = self._build_feedback(
                    final_reward_val, 
                    rollout.status, 
                    spans,
                    batch_inputs.get(rollout.rollout_id, {})
                )
                
                trajectories.append({
                    "rollout_id": rollout.rollout_id,
                    "original_input": batch_inputs.get(rollout.rollout_id, {}),
                    "assistant_response": assistant_response,
                    "feedback": feedback,
                    "spans": spans,
                })

        return EvaluationBatch(
            outputs=outputs,
            trajectories=trajectories,
            scores=scores,
        )

    def _extract_assistant_response(self, spans: List[Span]) -> str:
        """Extract the final assistant response from spans."""
        responses = []
        
        for span in spans:
            attrs = span.attributes if hasattr(span, 'attributes') else {}
            name = span.name if hasattr(span, 'name') else ""
            
            # Look for LLM completion spans
            if "chat" in name.lower() or "completion" in name.lower() or "llm" in name.lower():
                # Try different attribute keys for response content
                response_keys = [
                    "gen_ai.completion.0.content",
                    "gen_ai.response.content", 
                    "response.content",
                    "completion",
                ]
                for key in response_keys:
                    if key in attrs and attrs[key]:
                        responses.append(str(attrs[key]))
                        break
        
        if responses:
            return responses[-1]  # Return the last (final) response
        return "No assistant response found in spans"

    def _build_feedback(
        self, 
        reward: float, 
        status: str, 
        spans: List[Span],
        task_input: Dict[str, Any]
    ) -> str:
        """Build detailed feedback for reflection.
        
        For evaluation tasks, extracts human_score and llm_score from reward dimensions
        to provide specific feedback about prediction errors.
        """
        parts = []
        
        # Try to extract multi-dimensional reward data (human_score, llm_score)
        human_score = None
        llm_score = None
        
        reward_spans = find_reward_spans(spans)
        for span in reward_spans:
            rewards = get_rewards_from_span(span)
            for r in rewards:
                if r.name == "human_score":
                    human_score = r.value
                elif r.name == "llm_score":
                    llm_score = r.value
        
        # If we have both scores, provide detailed evaluation feedback
        if human_score is not None and llm_score is not None:
            diff = llm_score - human_score
            if abs(diff) < 0.1:
                parts.append(f"Good prediction: LLM scored {llm_score:.2f}, human scored {human_score:.2f} (close match).")
            elif diff > 0:
                parts.append(f"OVERESTIMATED: LLM scored {llm_score:.2f} but human scored {human_score:.2f}. "
                           f"The prompt caused the LLM to rate this {diff:.2f} points too high. "
                           "Consider adding stricter criteria or examples.")
            else:
                parts.append(f"UNDERESTIMATED: LLM scored {llm_score:.2f} but human scored {human_score:.2f}. "
                           f"The prompt caused the LLM to rate this {abs(diff):.2f} points too low. "
                           "Consider broadening the criteria or recognizing subtle details.")
        else:
            # Fallback to generic feedback
            if reward >= 0.9:
                parts.append(f"The task was completed successfully with a high score of {reward:.2f}.")
            elif reward >= 0.5:
                parts.append(f"The task was partially successful with a score of {reward:.2f}. There is room for improvement.")
            else:
                parts.append(f"The task failed or performed poorly with a score of {reward:.2f}. The prompt needs significant improvement.")
        
        # Add status info
        if status != "succeeded":
            parts.append(f"Rollout status: {status}.")
        
        # Look for error information in spans
        for span in spans:
            attrs = span.attributes if hasattr(span, 'attributes') else {}
            if "error" in str(attrs).lower():
                parts.append(f"Error detected in execution.")
                break
        
        # Add task context
        if task_input:
            if "expected_choice" in task_input:
                parts.append(f"Expected answer: {task_input['expected_choice']}")
            # Include humanScore from task if available (for debugging)
            if "humanScore" in task_input:
                parts.append(f"Expected human score: {task_input['humanScore']:.2f}")
        
        return " ".join(parts)

    def make_reflective_dataset(
        self,
        candidate: Dict[str, str],
        eval_batch: EvaluationBatch[MantisdkTrajectory, MantisdkRolloutOutput],
        components_to_update: List[str],
    ) -> Mapping[str, Sequence[Mapping[str, Any]]]:
        """Create a reflective dataset for GEPA's reflection mechanism.
        
        Returns data in GEPA's expected format:
        {
            "component_name": [
                {
                    "Inputs": str,           # Task input description
                    "Generated Outputs": str, # Model outputs
                    "Feedback": str           # Performance feedback
                },
                ...
            ]
        }
        """
        reflective_data: Dict[str, List[MantisdkReflectiveRecord]] = {
            component: [] for component in components_to_update
        }

        trajectories = eval_batch.trajectories
        if trajectories is None:
            logger.warning("No trajectories available for building reflective dataset")
            return reflective_data

        for i, traj in enumerate(trajectories):
            output = eval_batch.outputs[i]
            
            # Format the input as a readable string
            input_str = self._format_input(traj["original_input"])
            
            # Get the assistant response
            generated_output = traj["assistant_response"]
            
            # Get the feedback
            feedback = traj["feedback"]
            
            record: MantisdkReflectiveRecord = {
                "Inputs": input_str,
                "Generated Outputs": generated_output,
                "Feedback": feedback,
            }
            
            # Add to each component being updated
            for component in components_to_update:
                reflective_data[component].append(record)

        # Validate we have data
        for component in components_to_update:
            if len(reflective_data[component]) == 0:
                logger.warning(f"No reflective records for component {component}")

        return reflective_data

    def _format_input(self, task_input: Dict[str, Any]) -> str:
        """Format task input as a readable string for reflection."""
        if not task_input:
            return "No input available"
        
        # If it's a RoomSelectionTask-like structure
        if "task_input" in task_input:
            inner = task_input["task_input"]
            parts = []
            if "date" in inner:
                parts.append(f"Date: {inner['date']}")
            if "time" in inner:
                parts.append(f"Time: {inner['time']}")
            if "duration_min" in inner:
                parts.append(f"Duration: {inner['duration_min']} minutes")
            if "attendees" in inner:
                parts.append(f"Attendees: {inner['attendees']}")
            if "needs" in inner:
                parts.append(f"Needs: {', '.join(inner['needs']) if inner['needs'] else 'none'}")
            if "accessible_required" in inner:
                parts.append(f"Accessible required: {inner['accessible_required']}")
            return "; ".join(parts) if parts else json.dumps(task_input)
        
        # For other dict structures, try to format nicely
        try:
            if len(task_input) <= 5:
                return "; ".join(f"{k}: {v}" for k, v in task_input.items())
            return json.dumps(task_input, indent=2)
        except Exception:
            return str(task_input)



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import threading
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple

from opentelemetry.sdk.trace import ReadableSpan

from mantisdk.types import (
    Attempt,
    AttemptedRollout,
    AttemptStatus,
    EnqueueRolloutRequest,
    NamedResources,
    ResourcesUpdate,
    Rollout,
    RolloutConfig,
    RolloutStatus,
    Span,
    TaskInput,
    Worker,
    WorkerStatus,
)

from .base import UNSET, LightningStore, LightningStoreCapabilities, LightningStoreStatistics, Unset


class LightningStoreThreaded(LightningStore):
    """Facade that delegates all store operations to a underlying store instance.

    The operations are guaranteed to be thread-safe.
    Make sure the threaded stores are instantiated before initializing the threads.
    """

    def __init__(self, store: LightningStore) -> None:
        super().__init__()  # watchdog relies on the underlying store
        self.store = store
        self._lock = threading.Lock()

    @property
    def capabilities(self) -> LightningStoreCapabilities:
        """Return the capabilities of the store."""
        capabilities = self.store.capabilities
        return {
            **capabilities,
            "async_safe": True,
            "thread_safe": True,
        }

    async def statistics(self) -> LightningStoreStatistics:
        """Return the statistics of the store."""
        with self._lock:
            return await self.store.statistics()

    def otlp_traces_endpoint(self) -> str:
        """Return the OTLP/HTTP traces endpoint of the underlying store."""
        return self.store.otlp_traces_endpoint()

    def get_otlp_headers(self) -> Dict[str, str]:
        """Return the OTLP authentication headers from the underlying store."""
        if hasattr(self.store, "get_otlp_headers"):
            return self.store.get_otlp_headers()
        return {}

    async def start_rollout(
        self,
        input: TaskInput,
        mode: Literal["train", "val", "test"] | None = None,
        resources_id: str | None = None,
        config: RolloutConfig | None = None,
        metadata: Dict[str, Any] | None = None,
        worker_id: Optional[str] = None,
    ) -> AttemptedRollout:
        with self._lock:
            return await self.store.start_rollout(
                input,
                mode,
                resources_id,
                config,
                metadata,
                worker_id,
            )

    async def enqueue_rollout(
        self,
        input: TaskInput,
        mode: Literal["train", "val", "test"] | None = None,
        resources_id: str | None = None,
        config: RolloutConfig | None = None,
        metadata: Dict[str, Any] | None = None,
    ) -> Rollout:
        with self._lock:
            return await self.store.enqueue_rollout(input, mode, resources_id, config, metadata)

    async def enqueue_many_rollouts(self, rollouts: Sequence[EnqueueRolloutRequest]) -> Sequence[Rollout]:
        with self._lock:
            return await self.store.enqueue_many_rollouts(rollouts)

    async def dequeue_rollout(self, worker_id: Optional[str] = None) -> Optional[AttemptedRollout]:
        with self._lock:
            return await self.store.dequeue_rollout(worker_id=worker_id)

    async def dequeue_many_rollouts(
        self,
        *,
        limit: int = 1,
        worker_id: Optional[str] = None,
    ) -> Sequence[AttemptedRollout]:
        with self._lock:
            return await self.store.dequeue_many_rollouts(limit=limit, worker_id=worker_id)

    async def start_attempt(self, rollout_id: str, worker_id: Optional[str] = None) -> AttemptedRollout:
        with self._lock:
            return await self.store.start_attempt(rollout_id, worker_id)

    async def query_rollouts(
        self,
        *,
        status_in: Optional[Sequence[RolloutStatus]] = None,
        rollout_id_in: Optional[Sequence[str]] = None,
        rollout_id_contains: Optional[str] = None,
        filter_logic: Literal["and", "or"] = "and",
        sort_by: Optional[str] = None,
        sort_order: Literal["asc", "desc"] = "asc",
        limit: int = -1,
        offset: int = 0,
        status: Optional[Sequence[RolloutStatus]] = None,
        rollout_ids: Optional[Sequence[str]] = None,
    ) -> Sequence[Rollout]:
        with self._lock:
            return await self.store.query_rollouts(
                status_in=status_in,
                rollout_id_in=rollout_id_in,
                rollout_id_contains=rollout_id_contains,
                filter_logic=filter_logic,
                sort_by=sort_by,
                sort_order=sort_order,
                limit=limit,
                offset=offset,
                status=status,
                rollout_ids=rollout_ids,
            )

    async def query_attempts(
        self,
        rollout_id: str,
        *,
        sort_by: Optional[str] = "sequence_id",
        sort_order: Literal["asc", "desc"] = "asc",
        limit: int = -1,
        offset: int = 0,
    ) -> Sequence[Attempt]:
        with self._lock:
            return await self.store.query_attempts(
                rollout_id,
                sort_by=sort_by,
                sort_order=sort_order,
                limit=limit,
                offset=offset,
            )

    async def get_rollout_by_id(self, rollout_id: str) -> Optional[Rollout]:
        with self._lock:
            return await self.store.get_rollout_by_id(rollout_id)

    async def get_latest_attempt(self, rollout_id: str) -> Optional[Attempt]:
        with self._lock:
            return await self.store.get_latest_attempt(rollout_id)

    async def query_resources(
        self,
        *,
        resources_id: Optional[str] = None,
        resources_id_contains: Optional[str] = None,
        sort_by: Optional[str] = None,
        sort_order: Literal["asc", "desc"] = "asc",
        limit: int = -1,
        offset: int = 0,
    ) -> Sequence[ResourcesUpdate]:
        with self._lock:
            return await self.store.query_resources(
                resources_id=resources_id,
                resources_id_contains=resources_id_contains,
                sort_by=sort_by,
                sort_order=sort_order,
                limit=limit,
                offset=offset,
            )

    async def add_resources(self, resources: NamedResources) -> ResourcesUpdate:
        with self._lock:
            return await self.store.add_resources(resources)

    async def update_resources(self, resources_id: str, resources: NamedResources) -> ResourcesUpdate:
        with self._lock:
            return await self.store.update_resources(resources_id, resources)

    async def get_resources_by_id(self, resources_id: str) -> Optional[ResourcesUpdate]:
        with self._lock:
            return await self.store.get_resources_by_id(resources_id)

    async def get_latest_resources(self) -> Optional[ResourcesUpdate]:
        with self._lock:
            return await self.store.get_latest_resources()

    async def add_many_spans(self, spans: Sequence[Span]) -> Sequence[Span]:
        with self._lock:
            return await self.store.add_many_spans(spans)

    async def add_span(self, span: Span) -> Optional[Span]:
        with self._lock:
            return await self.store.add_span(span)

    async def add_otel_span(
        self,
        rollout_id: str,
        attempt_id: str,
        readable_span: ReadableSpan,
        sequence_id: int | None = None,
    ) -> Optional[Span]:
        with self._lock:
            return await self.store.add_otel_span(rollout_id, attempt_id, readable_span, sequence_id)

    async def wait_for_rollouts(self, *, rollout_ids: List[str], timeout: Optional[float] = None) -> List[Rollout]:
        # This method does not change the state of the store, and it's not thread-safe.
        return await self.store.wait_for_rollouts(rollout_ids=rollout_ids, timeout=timeout)

    async def get_next_span_sequence_id(self, rollout_id: str, attempt_id: str) -> int:
        with self._lock:
            return await self.store.get_next_span_sequence_id(rollout_id, attempt_id)

    async def get_many_span_sequence_ids(self, rollout_attempt_ids: Sequence[Tuple[str, str]]) -> Sequence[int]:
        with self._lock:
            return await self.store.get_many_span_sequence_ids(rollout_attempt_ids)

    async def query_spans(
        self,
        rollout_id: str,
        attempt_id: str | Literal["latest"] | None = None,
        *,
        trace_id: Optional[str] = None,
        trace_id_contains: Optional[str] = None,
        span_id: Optional[str] = None,
        span_id_contains: Optional[str] = None,
        parent_id: Optional[str] = None,
        parent_id_contains: Optional[str] = None,
        name: Optional[str] = None,
        name_contains: Optional[str] = None,
        filter_logic: Literal["and", "or"] = "and",
        limit: int = -1,
        offset: int = 0,
        sort_by: Optional[str] = "sequence_id",
        sort_order: Literal["asc", "desc"] = "asc",
    ) -> Sequence[Span]:
        with self._lock:
            return await self.store.query_spans(
                rollout_id,
                attempt_id,
                trace_id=trace_id,
                trace_id_contains=trace_id_contains,
                span_id=span_id,
                span_id_contains=span_id_contains,
                parent_id=parent_id,
                parent_id_contains=parent_id_contains,
                name=name,
                name_contains=name_contains,
                filter_logic=filter_logic,
                limit=limit,
                offset=offset,
                sort_by=sort_by,
                sort_order=sort_order,
            )

    async def update_rollout(
        self,
        rollout_id: str,
        input: TaskInput | Unset = UNSET,
        mode: Optional[Literal["train", "val", "test"]] | Unset = UNSET,
        resources_id: Optional[str] | Unset = UNSET,
        status: RolloutStatus | Unset = UNSET,
        config: RolloutConfig | Unset = UNSET,
        metadata: Optional[Dict[str, Any]] | Unset = UNSET,
    ) -> Rollout:
        with self._lock:
            return await self.store.update_rollout(
                rollout_id=rollout_id,
                input=input,
                mode=mode,
                resources_id=resources_id,
                status=status,
                config=config,
                metadata=metadata,
            )

    async def update_attempt(
        self,
        rollout_id: str,
        attempt_id: str | Literal["latest"],
        status: AttemptStatus | Unset = UNSET,
        worker_id: str | Unset = UNSET,
        last_heartbeat_time: float | Unset = UNSET,
        metadata: Optional[Dict[str, Any]] | Unset = UNSET,
    ) -> Attempt:
        with self._lock:
            return await self.store.update_attempt(
                rollout_id=rollout_id,
                attempt_id=attempt_id,
                status=status,
                worker_id=worker_id,
                last_heartbeat_time=last_heartbeat_time,
                metadata=metadata,
            )

    async def query_workers(
        self,
        *,
        status_in: Optional[Sequence[WorkerStatus]] = None,
        worker_id_contains: Optional[str] = None,
        filter_logic: Literal["and", "or"] = "and",
        sort_by: Optional[str] = None,
        sort_order: Literal["asc", "desc"] = "asc",
        limit: int = -1,
        offset: int = 0,
    ) -> Sequence[Worker]:
        with self._lock:
            return await self.store.query_workers(
                status_in=status_in,
                worker_id_contains=worker_id_contains,
                sort_by=sort_by,
                sort_order=sort_order,
                limit=limit,
                offset=offset,
            )

    async def get_worker_by_id(self, worker_id: str) -> Optional[Worker]:
        with self._lock:
            return await self.store.get_worker_by_id(worker_id)

    async def update_worker(
        self,
        worker_id: str,
        heartbeat_stats: Dict[str, Any] | Unset = UNSET,
    ) -> Worker:
        with self._lock:
            return await self.store.update_worker(
                worker_id=worker_id,
                heartbeat_stats=heartbeat_stats,
            )

    def complete_job(self, summary: Optional[Dict[str, Any]] = None) -> None:
        """Delegate job completion to the underlying store."""
        with self._lock:
            # We check if the underlying store has the method to be safe,
            # though base LightningStore now has it.
            if hasattr(self.store, "complete_job"):
                self.store.complete_job(summary)



# Copyright (c) Microsoft. All rights reserved.

import logging
from typing import Any, Dict, Optional

from mantisdk.semconv import AGL_MESSAGE, LightningSpanAttributes
from mantisdk.tracer.base import get_active_tracer
from mantisdk.tracer.dummy import DummyTracer
from mantisdk.types import Attributes, SpanLike
from mantisdk.utils.otel import flatten_attributes, sanitize_attributes

logger = logging.getLogger(__name__)


def emit_message(message: str, attributes: Optional[Dict[str, Any]] = None, propagate: bool = True) -> None:
    """Emit a textual message as an OpenTelemetry span.

    Commonly used for sending debugging and logging messages.

    Args:
        message: Human readable message to attach as a span attribute.
        attributes: Additional attributes to attach to the message span.
        propagate: Whether to propagate the span to exporters automatically.

    !!! note
        OpenTelemetry distinguishes between logs and spans. Emitting the message as a
        span keeps all Mantisdk telemetry in a single data store for analysis.
    """
    if not isinstance(message, str):  # type: ignore
        raise TypeError(f"Message must be a string or list of strings, got: {type(message)}.")

    if propagate:
        tracer = get_active_tracer()
        if tracer is None:
            raise RuntimeError("No active tracer found. Cannot emit message span.")
    else:
        tracer = DummyTracer()
    span_attributes: Attributes = {LightningSpanAttributes.MESSAGE_BODY.value: message}
    if attributes:
        flattened = flatten_attributes(attributes, expand_leaf_lists=False)
        span_attributes.update(sanitize_attributes(flattened))
    logger.debug("Emitting message span with message: %s", message)
    tracer.create_span(
        AGL_MESSAGE,
        attributes=span_attributes,
    )


def get_message_value(span: SpanLike) -> Optional[str]:
    """Extract the message string from a message span.

    Args:
        span: Span-like object to extract the message from.
    """
    span_attributes = span.attributes or {}
    if LightningSpanAttributes.MESSAGE_BODY.value not in span_attributes:
        return None
    message = span_attributes[LightningSpanAttributes.MESSAGE_BODY.value]
    if isinstance(message, str):
        return message
    raise TypeError(f"Message must be a string, got: {type(message)}.")



# Copyright (c) Microsoft. All rights reserved.

"""LiteLLM instrumentations.

Patches LiteLLM's OpenTelemetry integration to add Mantisdk-specific attributes
to spans, including session_id, tags, environment, and call_type from context.

The call_type is read from a ContextVar that is set by algorithm-specific
decorators (e.g., @gepa.judge, @gepa.agent).

[Related documentation](https://docs.litellm.ai/docs/observability/agentops_integration).
"""

import json
from typing import Any, Optional

from litellm.integrations.opentelemetry import OpenTelemetry

from mantisdk.types.tracing import get_current_call_type

__all__ = [
    "instrument_litellm",
    "uninstrument_litellm",
]

original_set_attributes = OpenTelemetry.set_attributes  # type: ignore


def patched_set_attributes(self: Any, span: Any, kwargs: Any, response_obj: Optional[Any]):
    """Enhanced set_attributes that adds Mantisdk tracing metadata to spans.
    
    Extracts session_id, tags, and environment from kwargs["metadata"] and sets
    them as OTEL span attributes for visibility in Insight/Langfuse.
    
    Also reads the current call_type from context (set by decorators like
    @gepa.judge) and adds it as a span attribute.
    """
    original_set_attributes(self, span, kwargs, response_obj)
    
    # Add token IDs if available
    if response_obj is not None and response_obj.get("prompt_token_ids"):
        span.set_attribute("prompt_token_ids", list(response_obj.get("prompt_token_ids")))
    if response_obj is not None and response_obj.get("response_token_ids"):
        span.set_attribute("response_token_ids", list(response_obj.get("response_token_ids")[0]))
    
    # Read call_type from context (set by @gepa.judge, @gepa.agent, etc.)
    call_type = get_current_call_type()
    if call_type:
        span.set_attribute("mantis.call_type", call_type)
    
    # Extract Mantisdk tracing metadata from kwargs
    metadata = kwargs.get("metadata", {}) if kwargs else {}
    if not metadata:
        return
    
    # Set session_id as span attribute (standard Langfuse/Insight attribute)
    session_id = metadata.get("session_id")
    if session_id:
        span.set_attribute("session.id", session_id)
    
    # Set tags as span attribute (JSON array for Langfuse compatibility)
    tags = metadata.get("tags")
    if tags and isinstance(tags, list):
        # Langfuse expects tags as a JSON array string or individual attributes
        span.set_attribute("tags", json.dumps(tags))
        # Also set individual tag attributes for filtering
        for i, tag in enumerate(tags):
            span.set_attribute(f"tag.{i}", str(tag))
    
    # Set environment as span attribute
    environment = metadata.get("environment")
    if environment:
        span.set_attribute("environment", environment)


def instrument_litellm():
    """Instrument litellm to capture token IDs."""
    OpenTelemetry.set_attributes = patched_set_attributes


def uninstrument_litellm():
    """Uninstrument litellm to stop capturing token IDs."""
    OpenTelemetry.set_attributes = original_set_attributes



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import functools
import logging
import random
from collections.abc import Coroutine
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Concatenate,
    Iterator,
    List,
    Literal,
    Optional,
    ParamSpec,
    Sequence,
    TypeVar,
    overload,
)

from mantisdk.types import Dataset

if TYPE_CHECKING:
    from mantisdk.llm_proxy import LLMProxy
    from mantisdk.store.base import LightningStore

    from .base import Algorithm

T_task = TypeVar("T_task")
T_algo = TypeVar("T_algo", bound="Algorithm")

P = ParamSpec("P")
R = TypeVar("R")


logger = logging.getLogger(__name__)


def batch_iter_over_dataset(dataset: Dataset[T_task], batch_size: int) -> Iterator[Sequence[T_task]]:
    """
    Create an infinite iterator that yields batches from the dataset.

    When batch_size >= dataset size, yields the entire shuffled dataset repeatedly.
    When batch_size < dataset size, yields batches of the specified size, reshuffling
    after each complete pass through the dataset.

    Args:
        dataset: The dataset to iterate over.
        batch_size: The desired batch size.

    Yields:
        Sequences of tasks from the dataset. Each task appears at most once per epoch.
    """
    if batch_size >= len(dataset):
        while True:
            dataset_copy = [dataset[i] for i in range(len(dataset))]
            random.shuffle(dataset_copy)
            yield dataset_copy

    else:
        current_batch: List[int] = []
        while True:
            indices = list(range(len(dataset)))
            random.shuffle(indices)
            for index in indices:
                if index in current_batch:
                    continue
                current_batch.append(index)
                if len(current_batch) == batch_size:
                    yield [dataset[index] for index in current_batch]
                    current_batch = []


def with_store(
    func: Callable[Concatenate[T_algo, LightningStore, P], Coroutine[Any, Any, R]],
) -> Callable[Concatenate[T_algo, P], Coroutine[Any, Any, R]]:
    """Inject the algorithm's `LightningStore` into coroutine methods.

    The decorator calls `Algorithm.get_store()` once per invocation and passes the
    resulting store as an explicit argument to the wrapped coroutine. Decorated
    methods therefore receive the resolved store even when invoked by helper
    utilities rather than directly by the algorithm.

    Args:
        func: The coroutine that expects `(self, store, *args, **kwargs)`.

    Returns:
        A coroutine wrapper that automatically retrieves the store and forwards it
        to `func`.
    """

    @functools.wraps(func)
    async def wrapper(self: T_algo, *args: P.args, **kwargs: P.kwargs) -> R:
        store = self.get_store()
        return await func(self, store, *args, **kwargs)

    return wrapper


@overload
def with_llm_proxy(
    required: Literal[False] = False,
    auto_start: bool = True,
) -> Callable[
    [Callable[Concatenate[T_algo, Optional[LLMProxy], P], Coroutine[Any, Any, R]]],
    Callable[Concatenate[T_algo, P], Coroutine[Any, Any, R]],
]: ...


@overload
def with_llm_proxy(
    required: Literal[True],
    auto_start: bool = True,
) -> Callable[
    [Callable[Concatenate[T_algo, LLMProxy, P], Coroutine[Any, Any, R]]],
    Callable[Concatenate[T_algo, P], Coroutine[Any, Any, R]],
]: ...


def with_llm_proxy(
    required: bool = False,
    auto_start: bool = True,
) -> Callable[
    [Callable[..., Coroutine[Any, Any, Any]]],
    Callable[..., Coroutine[Any, Any, Any]],
]:
    """Resolve and optionally lifecycle-manage the configured LLM proxy.

    Args:
        required: When True, raises `ValueError` if the algorithm does not have an
            [`LLMProxy`][mantisdk.LLMProxy] set. When False, the wrapped coroutine receives
            `None` if no proxy is available.
        auto_start: When True, [`LLMProxy.start()`][mantisdk.LLMProxy.start] is invoked if the proxy is not
            already running before calling `func` and [`LLMProxy.stop()`][mantisdk.LLMProxy.stop] is
            called afterwards.

    Returns:
        A decorator that injects the [`LLMProxy`][mantisdk.LLMProxy] (or `None`) as the first
        argument after `self` and manages automatic startup/shutdown when requested.
    """

    def decorator(
        func: Callable[..., Coroutine[Any, Any, Any]],
    ) -> Callable[..., Coroutine[Any, Any, Any]]:
        @functools.wraps(func)
        async def wrapper(self: Algorithm, *args: Any, **kwargs: Any) -> Any:
            llm_proxy = self.get_llm_proxy()

            if required and llm_proxy is None:
                raise ValueError(
                    "LLM proxy is required but not configured. Call set_llm_proxy() before using this method."
                )

            auto_started = False
            if auto_start and llm_proxy is not None:
                if llm_proxy.is_running():
                    logger.info("Proxy is already running, skipping start")
                else:
                    logger.info("Starting proxy, managed by the algorithm")
                    await llm_proxy.start()
                    auto_started = True

            try:
                # At type level, overloads guarantee that if `required=True`
                # then `func` expects a non-optional LLMProxy.
                return await func(self, llm_proxy, *args, **kwargs)
            finally:
                if auto_started and llm_proxy is not None:
                    logger.info("Stopping proxy, managed by the algorithm")
                    await llm_proxy.stop()

        return wrapper

    return decorator



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

"""Typed representations of tunable resources shared between Mantisdk components."""

import inspect
import logging
from typing import (
    Annotated,
    Any,
    Dict,
    Literal,
    Optional,
    Union,
)

from pydantic import BaseModel, Field

from .core import AttemptedRollout

logger = logging.getLogger(__name__)


__all__ = [
    "Resource",
    "LLM",
    "ProxyLLM",
    "PromptTemplate",
    "ResourceUnion",
    "NamedResources",
    "ResourcesUpdate",
]


class Resource(BaseModel):
    """Base class for tunable resources distributed to executors."""

    resource_type: Any
    """Alias of the resource type."""


class LLM(Resource):
    """Resource that identifies an LLM endpoint and its configuration."""

    resource_type: Literal["llm"] = "llm"
    endpoint: str
    """The URL of the LLM API endpoint."""
    model: str
    """The identifier for the model to be used (e.g., 'gpt-4o')."""
    api_key: Optional[str] = None
    """Optional secret used to authenticate requests."""
    sampling_parameters: Dict[str, Any] = Field(default_factory=dict)
    """A dictionary of hyperparameters for model inference, such as temperature, top_p, etc."""

    def get_base_url(self, *args: Any, **kwargs: Any) -> str:
        """Return the base URL consumed by OpenAI-compatible clients.

        Users are encouraged to use `get_base_url(rollout_id, attempt_id)` to get
        the LLM endpoint instead of accessing `.endpoint` directly.
        """
        return self.endpoint


class ProxyLLM(LLM):
    """LLM resource that rewrites endpoints through [`LLMProxy`][mantisdk.LLMProxy].

    The proxy injects rollout- and attempt-specific routing information into the
    endpoint so that downstream services can attribute requests correctly.
    """

    resource_type: Literal["proxy_llm"] = "proxy_llm"  # type: ignore
    _initialized: bool = False

    def model_post_init(self, __context: Any) -> None:
        """Mark initialization as complete after Pydantic finishes setup."""
        super().model_post_init(__context)
        object.__setattr__(self, "_initialized", True)

    def __getattribute__(self, name: str) -> Any:
        """Emit a warning when `endpoint` is accessed directly after initialization."""
        # Check if we're accessing endpoint after initialization and not from base_url
        if name == "endpoint":
            try:
                initialized = object.__getattribute__(self, "_initialized")
            except AttributeError:
                initialized = False

            if initialized:
                # Check the call stack to see if we're being called from base_url
                frame = inspect.currentframe()
                if frame and frame.f_back:
                    caller_name = frame.f_back.f_code.co_name
                    if caller_name != "get_base_url":
                        logger.warning(
                            "Accessing 'endpoint' directly on ProxyLLM is discouraged. "
                            "Use 'get_base_url(rollout_id, attempt_id)' instead to get the properly formatted endpoint."
                        )
        return super().__getattribute__(name)

    def with_attempted_rollout(self, rollout: AttemptedRollout) -> LLM:
        """Bake rollout metadata into a concrete [`LLM`][mantisdk.LLM] instance."""
        return LLM(
            endpoint=self.get_base_url(rollout.rollout_id, rollout.attempt.attempt_id),
            model=self.model,
            sampling_parameters=self.sampling_parameters,
            api_key=self.api_key,
        )

    def get_base_url(self, rollout_id: Optional[str], attempt_id: Optional[str]) -> str:
        """Return the routed endpoint for a specific rollout/attempt pair.

        Args:
            rollout_id: Identifier of the rollout making the request.
            attempt_id: Identifier of the attempt within that rollout.

        Returns:
            Fully qualified endpoint including rollout metadata.

        Raises:
            ValueError: If exactly one of ``rollout_id`` or ``attempt_id`` is provided.
        """
        if rollout_id is None and attempt_id is None:
            return self.endpoint

        if not (isinstance(rollout_id, str) and isinstance(attempt_id, str)):
            raise ValueError("rollout_id and attempt_id must be strings or all be empty")

        prefix = self.endpoint
        if prefix.endswith("/"):
            prefix = prefix[:-1]
        if prefix.endswith("/v1"):
            prefix = prefix[:-3]
            has_v1 = True
        else:
            has_v1 = False
        # Now the prefix should look like "http://localhost:11434"

        # Append the rollout and attempt id to the prefix
        prefix = prefix + f"/rollout/{rollout_id}/attempt/{attempt_id}"
        if has_v1:
            prefix = prefix + "/v1"
        return prefix


class PromptTemplate(Resource):
    """Resource describing a reusable prompt template."""

    resource_type: Literal["prompt_template"] = "prompt_template"
    template: str
    """The template string. The format depends on the engine."""
    engine: Literal["jinja", "f-string", "poml"]
    """The templating engine to use for rendering the prompt."""

    def format(self, **kwargs: Any) -> str:
        """Format the prompt using keyword arguments.

        !!! warning
            Only the `f-string` engine is supported for now.
        """
        if self.engine == "f-string":
            return self.template.format(**kwargs)
        else:
            raise NotImplementedError(
                "Formatting prompt templates for non-f-string engines with format() helper is not supported yet."
            )


# Use discriminated union for proper deserialization
# TODO: migrate to use a registry
ResourceUnion = Annotated[Union[LLM, ProxyLLM, PromptTemplate], Field(discriminator="resource_type")]
NamedResources = Dict[str, ResourceUnion]
"""Mapping from resource names to their configured instances.

Examples:
    ```python
    resources: NamedResources = {
        "main_llm": LLM(
            endpoint="http://localhost:8080",
            model="llama3",
            sampling_parameters={"temperature": 0.7, "max_tokens": 100},
        ),
        "system_prompt": PromptTemplate(
            template="You are a helpful assistant.",
            engine="f-string",
        ),
    }
    ```
"""


class ResourcesUpdate(BaseModel):
    """Update payload broadcast to clients when resources change."""

    resources_id: str
    """Identifier used to version the resources."""
    create_time: float
    """Timestamp of the creation time of the resources."""
    update_time: float
    """Timestamp of the last update time of the resources."""
    version: int
    """Version of the resources."""
    resources: NamedResources
    """Mapping of resource names to their definitions."""



# Copyright (c) 2025 Lakshya A Agrawal and the GEPA contributors
# https://github.com/gepa-ai/gepa

import random

from mantisdk.algorithm.gepa.lib.core.state import GEPAState
from mantisdk.algorithm.gepa.lib.gepa_utils import idxmax, select_program_candidate_from_pareto_front
from mantisdk.algorithm.gepa.lib.proposer.reflective_mutation.base import CandidateSelector


class ParetoCandidateSelector(CandidateSelector):
    def __init__(self, rng: random.Random | None):
        if rng is None:
            self.rng = random.Random(0)
        else:
            self.rng = rng

    def select_candidate_idx(self, state: GEPAState) -> int:
        assert len(state.program_full_scores_val_set) == len(state.program_candidates)
        return select_program_candidate_from_pareto_front(
            state.get_pareto_front_mapping(),
            state.per_program_tracked_scores,
            self.rng,
        )


class CurrentBestCandidateSelector(CandidateSelector):
    def __init__(self):
        pass

    def select_candidate_idx(self, state: GEPAState) -> int:
        assert len(state.program_full_scores_val_set) == len(state.program_candidates)
        return idxmax(state.program_full_scores_val_set)


class EpsilonGreedyCandidateSelector(CandidateSelector):
    def __init__(self, epsilon: float, rng: random.Random | None):
        assert 0.0 <= epsilon <= 1.0
        self.epsilon = epsilon
        if rng is None:
            self.rng = random.Random(0)
        else:
            self.rng = rng

    def select_candidate_idx(self, state: GEPAState) -> int:
        assert len(state.program_full_scores_val_set) == len(state.program_candidates)
        if self.rng.random() < self.epsilon:
            return self.rng.randint(0, len(state.program_candidates) - 1)
        else:
            return idxmax(state.program_full_scores_val_set)



# Copyright (c) Microsoft. All rights reserved.

# type: ignore

import torch
from datasets import Dataset as HuggingFaceDataset
from omegaconf import DictConfig
from verl.utils.dataset.rl_dataset import RLHFDataset

from mantisdk.types import Dataset

__all__ = [
    "AgentDataset",
    "LoadedDataset",
]


class AgentDataset(RLHFDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.filter_overlong_prompts = False

    def __getitem__(self, item):
        row_dict: dict = self.dataframe[item]

        # add index for each prompt
        index = row_dict.get("extra_info", {}).get("index", 0)
        row_dict["index"] = index
        # Workaround for data proto. At least one tensor is needed.
        row_dict["fake_ids"] = torch.ones(1, dtype=torch.int)
        return row_dict


class LoadedDataset(AgentDataset):

    def __init__(self, dataset: Dataset):
        super().__init__([], None, DictConfig({}))  # type: ignore
        dataset_copy = [dataset[i] for i in range(len(dataset))]
        self.dataframe = HuggingFaceDataset.from_list(dataset_copy)

    def _read_files_and_tokenize(self):
        pass



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import logging
import threading
import warnings
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Iterator, List

import weave.trace.weave_init
from pydantic import validate_call
from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.ids import generate_id
from weave.trace_server_bindings.client_interface import TraceServerClientInterface
from weave.trace_server_bindings.models import ServerInfoRes

logger = logging.getLogger(__name__)

__all__ = [
    "instrument_weave",
    "uninstrument_weave",
    "InMemoryWeaveTraceServer",
]


class InMemoryWeaveTraceServer(TraceServerClientInterface):
    """A minimal in-memory implementation of the TraceServerInterface.

    It stores calls and objects in local dictionaries and returns valid Pydantic
    responses to satisfy the Weave client and FullTraceServerInterface protocol.
    """

    def __init__(self):
        # Minimal storage to allow basic querying in tests
        self.calls: Dict[str, tsi.CallSchema] = {}
        self.partial_calls: Dict[str, Dict[str, Any]] = {}
        self.objs: Dict[str, Any] = {}
        self.files: Dict[str, bytes] = {}
        self.feedback: List[tsi.FeedbackCreateReq] = []

        self._call_threading_lock = threading.Lock()

    @classmethod
    def from_env(cls, *args: Any, **kwargs: Any) -> InMemoryWeaveTraceServer:
        return cls()

    def server_info(self) -> ServerInfoRes:
        return ServerInfoRes(min_required_weave_python_version="0.52.22")

    def ensure_project_exists(self, entity: str, project: str) -> tsi.EnsureProjectExistsRes:
        return tsi.EnsureProjectExistsRes(project_name=project)

    # --- Call API ---

    @validate_call
    def call_start(self, req: tsi.CallStartReq) -> tsi.CallStartRes:
        # NOTE: It's not necessary that call_end must be called after call_start.
        request_content = req.start.model_dump(exclude_none=True)

        # If id needs to be generated here, it's very likely we won't be able to find the call later.
        # This is just to make the type checker happy.
        call_id = request_content.get("id") or generate_id()
        trace_id = request_content.get("trace_id") or generate_id()
        request_content["id"] = call_id
        request_content["trace_id"] = trace_id

        with self._call_threading_lock:
            if call_id in self.partial_calls:
                # call_end has already been called for this call.
                kwargs = {**request_content, **self.partial_calls[call_id]}
                self.calls[call_id] = tsi.CallSchema(**kwargs)
                del self.partial_calls[call_id]
            else:
                self.partial_calls[call_id] = request_content

        return tsi.CallStartRes(id=call_id, trace_id=trace_id)

    @validate_call
    def call_end(self, req: tsi.CallEndReq) -> tsi.CallEndRes:
        request_content = req.end.model_dump(exclude_none=True)
        call_id = req.end.id

        with self._call_threading_lock:
            if call_id in self.partial_calls:
                # End request always override the start request content.
                kwargs = {**self.partial_calls[call_id], **request_content}
                self.calls[call_id] = tsi.CallSchema(**kwargs)
                del self.partial_calls[call_id]
            else:
                self.partial_calls[call_id] = request_content
        return tsi.CallEndRes()

    @validate_call
    def call_start_batch(self, req: tsi.CallCreateBatchReq) -> tsi.CallCreateBatchRes:
        for item in req.batch:
            if isinstance(item, tsi.CallStartReq):
                self.call_start(item)
            elif isinstance(item, tsi.CallEndReq):
                self.call_end(item)
        return tsi.CallCreateBatchRes(res=[])

    @validate_call
    def call_read(self, req: tsi.CallReadReq) -> tsi.CallReadRes:
        call_data = self.calls.get(req.id)
        return tsi.CallReadRes(call=call_data)

    @validate_call
    def calls_query(self, req: tsi.CallsQueryReq) -> tsi.CallsQueryRes:
        return tsi.CallsQueryRes(calls=list(self.calls_query_stream(req)))

    @validate_call
    def calls_query_stream(self, req: tsi.CallsQueryReq) -> Iterator[tsi.CallSchema]:
        yield from self.calls.values()

    @validate_call
    def calls_delete(self, req: tsi.CallsDeleteReq) -> tsi.CallsDeleteRes:
        num_deleted = 0
        for call_id in req.call_ids:
            if call_id in self.calls:
                del self.calls[call_id]
                num_deleted += 1
        return tsi.CallsDeleteRes(num_deleted=num_deleted)

    @validate_call
    def call_update(self, req: tsi.CallUpdateReq) -> tsi.CallUpdateRes:
        return tsi.CallUpdateRes()

    @validate_call
    def calls_query_stats(self, req: tsi.CallsQueryStatsReq) -> tsi.CallsQueryStatsRes:
        return tsi.CallsQueryStatsRes(count=len(self.calls))

    # --- Cost API ---

    @validate_call
    def cost_create(self, req: tsi.CostCreateReq) -> tsi.CostCreateRes:
        return tsi.CostCreateRes(ids=[(generate_id(), generate_id()) for _ in req.costs])

    @validate_call
    def cost_query(self, req: tsi.CostQueryReq) -> tsi.CostQueryRes:
        return tsi.CostQueryRes(results=[])

    @validate_call
    def cost_purge(self, req: tsi.CostPurgeReq) -> tsi.CostPurgeRes:
        return tsi.CostPurgeRes()

    # --- Object API (Legacy V1) ---

    @validate_call
    def obj_create(self, req: tsi.ObjCreateReq) -> tsi.ObjCreateRes:
        digest = generate_id()
        self.objs[digest] = req.obj
        return tsi.ObjCreateRes(digest=digest)

    @validate_call
    def obj_read(self, req: tsi.ObjReadReq) -> tsi.ObjReadRes:
        return tsi.ObjReadRes(obj=self.objs.get(req.digest, {}))

    @validate_call
    def objs_query(self, req: tsi.ObjQueryReq) -> tsi.ObjQueryRes:
        return tsi.ObjQueryRes(objs=[])

    @validate_call
    def obj_delete(self, req: tsi.ObjDeleteReq) -> tsi.ObjDeleteRes:
        return tsi.ObjDeleteRes(num_deleted=0)

    # --- Table API ---

    @validate_call
    def table_create(self, req: tsi.TableCreateReq) -> tsi.TableCreateRes:
        return tsi.TableCreateRes(digest=generate_id(), row_digests=[])

    @validate_call
    def table_create_from_digests(self, req: tsi.TableCreateFromDigestsReq) -> tsi.TableCreateFromDigestsRes:
        return tsi.TableCreateFromDigestsRes(digest=generate_id())

    @validate_call
    def table_update(self, req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        return tsi.TableUpdateRes(digest=generate_id(), updated_row_digests=[])

    @validate_call
    def table_query(self, req: tsi.TableQueryReq) -> tsi.TableQueryRes:
        return tsi.TableQueryRes(rows=[])

    @validate_call
    def table_query_stream(self, req: tsi.TableQueryReq) -> Iterator[tsi.TableRowSchema]:
        yield from []

    @validate_call
    def table_query_stats(self, req: tsi.TableQueryStatsReq) -> tsi.TableQueryStatsRes:
        return tsi.TableQueryStatsRes(count=0)

    @validate_call
    def table_query_stats_batch(self, req: tsi.TableQueryStatsBatchReq) -> tsi.TableQueryStatsBatchRes:
        return tsi.TableQueryStatsBatchRes(tables=[])

    # --- Ref API ---

    @validate_call
    def refs_read_batch(self, req: tsi.RefsReadBatchReq) -> tsi.RefsReadBatchRes:
        return tsi.RefsReadBatchRes(vals=[])

    # --- File API ---

    def file_create(self, req: tsi.FileCreateReq) -> tsi.FileCreateRes:
        self.files[req.name] = req.content
        return tsi.FileCreateRes(digest=generate_id())

    def file_content_read(self, req: tsi.FileContentReadReq) -> tsi.FileContentReadRes:
        return tsi.FileContentReadRes(content=self.files.get(req.digest, b"dummy_content"))

    def files_stats(self, req: tsi.FilesStatsReq) -> tsi.FilesStatsRes:
        total_size = sum(len(c) for c in self.files.values())
        return tsi.FilesStatsRes(total_size_bytes=total_size)

    # --- Feedback API ---

    @validate_call
    def feedback_create(self, req: tsi.FeedbackCreateReq) -> tsi.FeedbackCreateRes:
        req.id = req.id or generate_id()
        self.feedback.append(req)
        return tsi.FeedbackCreateRes(
            id=req.id,
            created_at=datetime.now(timezone.utc),
            wb_user_id="dummy_user",
            payload=req.payload,
        )

    def feedback_create_batch(self, req: tsi.FeedbackCreateBatchReq) -> tsi.FeedbackCreateBatchRes:
        results: List[tsi.FeedbackCreateRes] = []
        for item in req.batch:
            res = self.feedback_create(item)
            results.append(res)
        return tsi.FeedbackCreateBatchRes(res=results)

    @validate_call
    def feedback_query(self, req: tsi.FeedbackQueryReq) -> tsi.FeedbackQueryRes:
        return tsi.FeedbackQueryRes(result=[])

    @validate_call
    def feedback_purge(self, req: tsi.FeedbackPurgeReq) -> tsi.FeedbackPurgeRes:
        self.feedback.clear()
        return tsi.FeedbackPurgeRes()

    @validate_call
    def feedback_replace(self, req: tsi.FeedbackReplaceReq) -> tsi.FeedbackReplaceRes:
        return tsi.FeedbackReplaceRes(
            id=req.id or generate_id(),
            created_at=datetime.now(timezone.utc),
            wb_user_id="dummy",
            payload={},
        )

    # --- Action API ---

    @validate_call
    def actions_execute_batch(self, req: tsi.ActionsExecuteBatchReq) -> tsi.ActionsExecuteBatchRes:
        return tsi.ActionsExecuteBatchRes()

    # --- Execute LLM API ---

    @validate_call
    def completions_create(self, req: tsi.CompletionsCreateReq) -> tsi.CompletionsCreateRes:
        return tsi.CompletionsCreateRes(response={"choices": [{"text": "dummy completion"}]})

    @validate_call
    def completions_create_stream(self, req: tsi.CompletionsCreateReq) -> Iterator[dict[str, Any]]:
        yield {"choices": [{"text": "dummy "}]}
        yield {"choices": [{"text": "stream"}]}

    # --- Execute Image Generation API ---

    @validate_call
    def image_create(self, req: tsi.ImageGenerationCreateReq) -> tsi.ImageGenerationCreateRes:
        return tsi.ImageGenerationCreateRes(response={})

    # --- Project Statistics API ---

    @validate_call
    def project_stats(self, req: tsi.ProjectStatsReq) -> tsi.ProjectStatsRes:
        return tsi.ProjectStatsRes(
            trace_storage_size_bytes=0,
            objects_storage_size_bytes=0,
            tables_storage_size_bytes=0,
            files_storage_size_bytes=0,
        )

    # --- Thread API ---

    @validate_call
    def threads_query_stream(self, req: tsi.ThreadsQueryReq) -> Iterator[tsi.ThreadSchema]:
        yield from []

    # --- Evaluation API (V1) ---

    @validate_call
    def evaluate_model(self, req: tsi.EvaluateModelReq) -> tsi.EvaluateModelRes:
        return tsi.EvaluateModelRes(call_id=generate_id())

    @validate_call
    def evaluation_status(self, req: tsi.EvaluationStatusReq) -> tsi.EvaluationStatusRes:
        return tsi.EvaluationStatusRes(status=tsi.EvaluationStatusNotFound())

    # --- OTEL API ---

    def otel_export(self, req: tsi.OtelExportReq) -> tsi.OtelExportRes:
        return tsi.OtelExportRes()

    # ==========================================
    # Object Interface (V2 APIs)
    # ==========================================

    # --- Ops ---
    def op_create(self, req: tsi.OpCreateReq) -> tsi.OpCreateRes:
        return tsi.OpCreateRes(digest=generate_id(), object_id=generate_id(), version_index=0)

    def op_read(self, req: tsi.OpReadReq) -> tsi.OpReadRes:
        return tsi.OpReadRes(op=None)  # type: ignore

    def op_list(self, req: tsi.OpListReq) -> Iterator[tsi.OpReadRes]:
        yield from []

    def op_delete(self, req: tsi.OpDeleteReq) -> tsi.OpDeleteRes:
        return tsi.OpDeleteRes(num_deleted=0)

    # --- Datasets ---
    def dataset_create(self, req: tsi.DatasetCreateReq) -> tsi.DatasetCreateRes:
        return tsi.DatasetCreateRes(digest=generate_id(), object_id=generate_id(), version_index=0)

    def dataset_read(self, req: tsi.DatasetReadReq) -> tsi.DatasetReadRes:
        return tsi.DatasetReadRes(dataset=None)  # type: ignore

    def dataset_list(self, req: tsi.DatasetListReq) -> Iterator[tsi.DatasetReadRes]:
        yield from []

    def dataset_delete(self, req: tsi.DatasetDeleteReq) -> tsi.DatasetDeleteRes:
        return tsi.DatasetDeleteRes(num_deleted=0)

    # --- Scorers ---
    def scorer_create(self, req: tsi.ScorerCreateReq) -> tsi.ScorerCreateRes:
        return tsi.ScorerCreateRes(digest=generate_id(), object_id=generate_id(), version_index=0, scorer=generate_id())

    def scorer_read(self, req: tsi.ScorerReadReq) -> tsi.ScorerReadRes:
        return tsi.ScorerReadRes(scorer=None)  # type: ignore

    def scorer_list(self, req: tsi.ScorerListReq) -> Iterator[tsi.ScorerReadRes]:
        yield from []

    def scorer_delete(self, req: tsi.ScorerDeleteReq) -> tsi.ScorerDeleteRes:
        return tsi.ScorerDeleteRes(num_deleted=0)

    # --- Evaluations (V2) ---
    def evaluation_create(self, req: tsi.EvaluationCreateReq) -> tsi.EvaluationCreateRes:
        return tsi.EvaluationCreateRes(
            digest=generate_id(), object_id=generate_id(), version_index=0, evaluation_ref=generate_id()
        )

    def evaluation_read(self, req: tsi.EvaluationReadReq) -> tsi.EvaluationReadRes:
        return tsi.EvaluationReadRes(evaluation=None)  # type: ignore

    def evaluation_list(self, req: tsi.EvaluationListReq) -> Iterator[tsi.EvaluationReadRes]:
        yield from []

    def evaluation_delete(self, req: tsi.EvaluationDeleteReq) -> tsi.EvaluationDeleteRes:
        return tsi.EvaluationDeleteRes(num_deleted=0)

    # --- Models ---
    def model_create(self, req: tsi.ModelCreateReq) -> tsi.ModelCreateRes:
        return tsi.ModelCreateRes(
            digest=generate_id(), object_id=generate_id(), version_index=0, model_ref=generate_id()
        )

    def model_read(self, req: tsi.ModelReadReq) -> tsi.ModelReadRes:
        return tsi.ModelReadRes(model=None)  # type: ignore

    def model_list(self, req: tsi.ModelListReq) -> Iterator[tsi.ModelReadRes]:
        yield from []

    def model_delete(self, req: tsi.ModelDeleteReq) -> tsi.ModelDeleteRes:
        return tsi.ModelDeleteRes(num_deleted=0)

    # --- Evaluation Runs ---
    def evaluation_run_create(self, req: tsi.EvaluationRunCreateReq) -> tsi.EvaluationRunCreateRes:
        return tsi.EvaluationRunCreateRes(evaluation_run_id=generate_id())

    def evaluation_run_read(self, req: tsi.EvaluationRunReadReq) -> tsi.EvaluationRunReadRes:
        return tsi.EvaluationRunReadRes(evaluation_run=None)  # type: ignore

    def evaluation_run_list(self, req: tsi.EvaluationRunListReq) -> Iterator[tsi.EvaluationRunReadRes]:
        yield from []

    def evaluation_run_delete(self, req: tsi.EvaluationRunDeleteReq) -> tsi.EvaluationRunDeleteRes:
        return tsi.EvaluationRunDeleteRes(num_deleted=0)

    def evaluation_run_finish(self, req: tsi.EvaluationRunFinishReq) -> tsi.EvaluationRunFinishRes:
        return tsi.EvaluationRunFinishRes(success=True)

    # --- Predictions ---
    def prediction_create(self, req: tsi.PredictionCreateReq) -> tsi.PredictionCreateRes:
        return tsi.PredictionCreateRes(prediction_id=generate_id())

    def prediction_read(self, req: tsi.PredictionReadReq) -> tsi.PredictionReadRes:
        return tsi.PredictionReadRes(prediction=None)  # type: ignore

    def prediction_list(self, req: tsi.PredictionListReq) -> Iterator[tsi.PredictionReadRes]:
        yield from []

    def prediction_delete(self, req: tsi.PredictionDeleteReq) -> tsi.PredictionDeleteRes:
        return tsi.PredictionDeleteRes(num_deleted=0)

    def prediction_finish(self, req: tsi.PredictionFinishReq) -> tsi.PredictionFinishRes:
        return tsi.PredictionFinishRes(success=True)

    # --- Scores ---
    def score_create(self, req: tsi.ScoreCreateReq) -> tsi.ScoreCreateRes:
        return tsi.ScoreCreateRes(score_id=generate_id())

    def score_read(self, req: tsi.ScoreReadReq) -> tsi.ScoreReadRes:
        return tsi.ScoreReadRes(score=None)  # type: ignore

    def score_list(self, req: tsi.ScoreListReq) -> Iterator[tsi.ScoreReadRes]:
        yield from []

    def score_delete(self, req: tsi.ScoreDeleteReq) -> tsi.ScoreDeleteRes:
        return tsi.ScoreDeleteRes(num_deleted=0)


# Module-level storage for originals
_original_init_weave_get_server: Callable[..., Any] | None = None
_original_get_entity_project_from_project_name: Callable[..., Any] | None = None
_original_get_username: Callable[..., Any] | None = None


def init_weave_get_server_factory(server: InMemoryWeaveTraceServer) -> Callable[..., Any]:
    # Bypass the usage of Weave remote server
    def init_weave_get_server(*args: Any, **kwargs: Any) -> InMemoryWeaveTraceServer:
        return server

    return init_weave_get_server


def get_entity_project_from_project_name_factory(entity_name: str) -> tuple[str, str]:
    # Bypass the usage of API
    try:
        assert _original_get_entity_project_from_project_name is not None
        if _original_get_entity_project_from_project_name is not get_entity_project_from_project_name_factory:
            return _original_get_entity_project_from_project_name(entity_name)
        else:
            warnings.warn("W&B integration might have been repeatedly/recursively instrumented.")
            return "msk", "weave"
    except weave.trace.weave_init.WeaveWandbAuthenticationException:
        # In case API is not available.
        return "msk", "weave"


def get_username() -> str:
    # Bypass the usage of API
    try:
        assert _original_get_username is not None
        return _original_get_username()
    except RuntimeError:
        return "msk"
    except Exception as exc:
        warnings.warn(f"Unexpected error in get_username. Using default username. Error: {exc}")
        return "msk"


def instrument_weave(server: InMemoryWeaveTraceServer):
    """Patch the Weave/W&B integration to bypass actual network calls for testing."""

    global _original_init_weave_get_server, _original_get_entity_project_from_project_name, _original_get_username
    _original_init_weave_get_server = weave.trace.weave_init.init_weave_get_server
    _original_get_entity_project_from_project_name = weave.trace.weave_init.get_entity_project_from_project_name
    _original_get_username = weave.trace.weave_init.get_username
    weave.trace.weave_init.init_weave_get_server = init_weave_get_server_factory(server)
    weave.trace.weave_init.get_entity_project_from_project_name = get_entity_project_from_project_name_factory
    weave.trace.weave_init.get_username = get_username


def uninstrument_weave():
    """Restore the original Weave/W&B integration methods and HTTP requests."""
    global _original_init_weave_get_server, _original_get_entity_project_from_project_name, _original_get_username

    if _original_init_weave_get_server is not None:
        weave.trace.weave_init.init_weave_get_server = _original_init_weave_get_server
        _original_init_weave_get_server = None
    else:
        raise RuntimeError("Weave/W&B integration was not instrumented.")

    if _original_get_entity_project_from_project_name is not None:
        weave.trace.weave_init.get_entity_project_from_project_name = _original_get_entity_project_from_project_name
        _original_get_entity_project_from_project_name = None
    else:
        raise RuntimeError("Weave/W&B integration was not instrumented.")

    if _original_get_username is not None:
        weave.trace.weave_init.get_username = _original_get_username
        _original_get_username = None
    else:
        raise RuntimeError("Weave/W&B integration was not instrumented.")



# Copyright (c) Microsoft. All rights reserved.

from .base import Adapter, OtelTraceAdapter, TraceAdapter
from .messages import TraceToMessages
from .triplet import LlmProxyTraceToTriplet, TracerTraceToTriplet, TraceToTripletBase

__all__ = [
    "TraceAdapter",
    "OtelTraceAdapter",
    "Adapter",
    "TraceToTripletBase",
    "TracerTraceToTriplet",
    "LlmProxyTraceToTriplet",
    "TraceToMessages",
]



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import platform
import socket
from contextlib import suppress
from datetime import datetime
from typing import Any, Dict, List, cast

import psutil
from gpustat import GPUStat, GPUStatCollection


def system_snapshot(include_gpu: bool = False) -> Dict[str, Any]:
    """Capture a snapshot of the system's hardware and software information.

    Args:
        include_gpu: Whether to include GPU information.

    Returns:
        A dictionary containing the system's hardware and software information.
    """
    # CPU
    cpu = {
        "cpu_name": platform.processor(),
        "cpu_cores": psutil.cpu_count(logical=False),
        "cpu_threads": psutil.cpu_count(logical=True),
        "cpu_usage_pct": psutil.cpu_percent(0.0),
    }

    # Memory
    vm = psutil.virtual_memory()
    mem = {
        "mem_used_gb": round(vm.used / (2**30), 2),
        "mem_total_gb": round(vm.total / (2**30), 2),
        "mem_pct": vm.percent,
    }

    # Disk
    du = psutil.disk_usage("/")
    disk = {
        "disk_used_gb": round(du.used / (2**30), 2),
        "disk_total_gb": round(du.total / (2**30), 2),
        "disk_pct": du.percent,
    }

    # GPU (only query if explicitly requested)
    gpus: List[Dict[str, Any]] = []
    if include_gpu:
        with suppress(Exception):
            for g in GPUStatCollection.new_query().gpus:  # type: ignore
                g = cast(GPUStat, g)
                gpus.append(
                    {
                        "gpu": g.name,  # type: ignore
                        "util_pct": g.utilization,
                        "mem_used_mb": g.memory_used,
                        "mem_total_mb": g.memory_total,
                        "temp_c": g.temperature,
                    }
                )

    # Network
    net = psutil.net_io_counters()
    netinfo = {
        "bytes_sent_mb": round(net.bytes_sent / (2**20), 2),
        "bytes_recv_mb": round(net.bytes_recv / (2**20), 2),
    }

    # OS / meta
    return {
        "timestamp": datetime.now().isoformat(timespec="seconds"),
        "host": socket.gethostname(),
        "os": platform.platform(),
        **cpu,
        **mem,
        **disk,
        **netinfo,
        **({"gpus": gpus} if include_gpu else {}),
    }



# Copyright (c) Microsoft. All rights reserved.

import asyncio
import functools
import logging
import warnings
from typing import Any, Callable, Dict, Optional, Sequence, TypeVar, Union

from mantisdk.adapter import TraceAdapter, TracerTraceToTriplet
from mantisdk.algorithm import Algorithm, Baseline, FastAlgorithm
from mantisdk.client import MantisdkClient
from mantisdk.execution.base import ExecutionStrategy
from mantisdk.execution.client_server import ClientServerExecutionStrategy
from mantisdk.execution.events import ExecutionEvent
from mantisdk.litagent import LitAgent
from mantisdk.llm_proxy import LLMProxy
from mantisdk.runner import LitAgentRunner, Runner
from mantisdk.store.base import LightningStore
from mantisdk.store.memory import InMemoryLightningStore
from mantisdk.tracer.agentops import AgentOpsTracer
from mantisdk.tracer.base import Tracer
from mantisdk.tracer.otel import OtelTracer
from mantisdk.types import Dataset, Hook, NamedResources

from .init_utils import build_component, instantiate_component
from .legacy import TrainerLegacy
from .registry import ExecutionStrategyRegistry

logger = logging.getLogger(__name__)

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")

ComponentSpec = Union[T, type[T], Callable[[], T], str, Dict[str, Any], None]


class Trainer(TrainerLegacy):
    """High-level orchestration layer that wires Algorithm <-> Runner <-> Store.

    A [`Trainer`][mantisdk.Trainer] packages the moving parts of Agent-Lightning's
    training loop into a single entry point:

    * **Algorithm lifecycle:** Instantiates or accepts an [`Algorithm`][mantisdk.Algorithm],
      attaches the current [`LightningStore`][mantisdk.LightningStore], adapter, and
      initial resources, then executes the algorithm role inside the configured execution strategy.
    * **Runner fleet:** Spawns one or more [`Runner`][mantisdk.Runner] instances (defaulting
      to [`LitAgentRunner`][mantisdk.LitAgentRunner]) that hydrate a [`LitAgent`][mantisdk.LitAgent],
      claim rollouts, stream spans, and respect graceful termination signals from the execution strategy.
    * **Execution strategy:** Delegates process management to an
      [`ExecutionStrategy`][mantisdk.ExecutionStrategy] (shared memory, client/server, etc.),
      so advanced users can swap orchestration backends without changing trainer code.
    * **Telemetry plumbing:** Ensures tracers, adapters, and optional [`LLMProxy`][mantisdk.LLMProxy]
      are wired into both algorithm and runners so telemetry flows back into the store.

    The trainer exposes two convenience entry points:
    [`fit()`][mantisdk.Trainer.fit] for full training and
    [`dev()`][mantisdk.Trainer.dev] for fast, reproducible dry-runs. See the
    [Train the First Agent](../how-to/train-first-agent.md) and
    [Write the First Algorithm](../how-to/write-first-algorithm.md) tutorials for the broader context.
    """

    algorithm: Optional[Algorithm]
    """An instance of [`Algorithm`][mantisdk.Algorithm] to use for training."""

    store: LightningStore
    """An instance of [`LightningStore`][mantisdk.LightningStore] to use for storing tasks and traces."""

    runner: Runner[Any]
    """An instance of [`Runner`][mantisdk.Runner] to use for running the agent."""

    initial_resources: Optional[NamedResources]
    """An instance of [`NamedResources`][mantisdk.NamedResources] to use for bootstrapping the fit/dev process.

    The resources will be handed over to the algorithm. Note that not all algorithms support seeding resources.
    """

    n_runners: int
    """Number of agent runners to run in parallel."""

    max_rollouts: Optional[int]
    """Maximum number of rollouts to process per runner. If None, workers run until no more rollouts are available."""

    strategy: ExecutionStrategy
    """An instance of [`ExecutionStrategy`][mantisdk.ExecutionStrategy] to use for spawning the algorithm and runners."""

    tracer: Tracer
    """A tracer instance, or a string pointing to the class full name or a dictionary with a 'type' key
    that specifies the class full name and other initialization parameters.
    If None, a default [`AgentOpsTracer`][mantisdk.AgentOpsTracer] will be created with the current settings."""

    hooks: Sequence[Hook]
    """A sequence of [`Hook`][mantisdk.Hook] instances to be called at various lifecycle stages (e.g., `on_trace_start`,
    `on_trace_end`, `on_rollout_start`, `on_rollout_end`)."""

    adapter: TraceAdapter[Any]
    """An instance of [`TraceAdapter`][mantisdk.TraceAdapter] to export data consumble by algorithms from traces."""

    llm_proxy: Optional[LLMProxy]
    """An instance of [`LLMProxy`][mantisdk.LLMProxy] to use for intercepting the LLM calls.
    If not provided, algorithm may create one on its own."""

    n_workers: int
    """Number of agent workers to run in parallel. Deprecated in favor of `n_runners`."""

    max_tasks: Optional[int]
    """Maximum number of tasks to process per runner. Deprecated in favor of `max_rollouts`."""

    daemon: bool
    """Whether worker processes should be daemons. Daemon processes
    are terminated automatically when the main process exits. Deprecated.
    Only have effect with `fit_v0`."""

    triplet_exporter: TraceAdapter[Any]
    """An instance of [`TracerTraceToTriplet`][mantisdk.TracerTraceToTriplet] to export triplets from traces,
    or a dictionary with the initialization parameters for the exporter.
    Deprecated. Use [`adapter`][mantisdk.Trainer.adapter] instead."""

    port: Optional[int]
    """Port forwarded to [`ClientServerExecutionStrategy`][mantisdk.ClientServerExecutionStrategy]."""

    def __init__(
        self,
        *,
        dev: bool = False,
        n_runners: Optional[int] = None,
        max_rollouts: Optional[int] = None,
        initial_resources: Optional[NamedResources] = None,
        tracer: ComponentSpec[Tracer] = None,
        adapter: ComponentSpec[TraceAdapter[Any]] = None,
        store: ComponentSpec[LightningStore] = None,
        runner: ComponentSpec[Runner[Any]] = None,
        strategy: ComponentSpec[ExecutionStrategy] = None,
        port: Optional[int] = None,
        algorithm: ComponentSpec[Algorithm] = None,
        llm_proxy: ComponentSpec[LLMProxy] = None,
        n_workers: Optional[int] = None,
        max_tasks: Optional[int] = None,
        daemon: bool = True,
        triplet_exporter: ComponentSpec[TracerTraceToTriplet] = None,
        hooks: Optional[Union[Hook, Sequence[Hook]]] = None,
    ):
        """Configure the trainer and resolve user-provided component specifications.

        Each keyword accepts either a concrete instance, a class, a callable factory, a
        registry string, or a lightweight configuration dictionary (see
        [`build_component()`][mantisdk.trainer.init_utils.build_component]).

        When ``port`` is provided it is forwarded to
        [`ClientServerExecutionStrategy`][mantisdk.ClientServerExecutionStrategy]
        instances constructed (or supplied) for the trainer.
        """
        # Do not call super().__init__() here.
        # super().__init__() will call TrainerLegacy's initialization, which is not intended.
        self.worker_id: Optional[int] = None

        if dev:
            warnings.warn(
                "Trainer(dev=True) is deprecated and will be removed in future versions. "
                "Please use Trainer.dev(...) instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._dev = dev
        self.daemon = daemon
        self._client: MantisdkClient | None = None  # Will be initialized in fit or fit_v0

        if n_workers is not None:
            warnings.warn(
                "`n_workers` is deprecated. Please use `n_runners`.",
                DeprecationWarning,
                stacklevel=2,
            )

        if n_runners is None:
            n_runners = n_workers if n_workers is not None else 1
        else:
            if n_workers is not None and n_workers != n_runners:
                warnings.warn(
                    "`n_workers` is ignored when `n_runners` is provided.",
                    DeprecationWarning,
                    stacklevel=2,
                )

        self.n_runners = n_runners
        self.n_workers = n_runners  # Backwards compatibility for fit_v0

        if max_tasks is not None:
            warnings.warn(
                "`max_tasks` is deprecated. Please use `max_rollouts`.",
                DeprecationWarning,
                stacklevel=2,
            )

        if max_rollouts is None:
            max_rollouts = max_tasks
        elif max_tasks is not None and max_tasks != max_rollouts:
            warnings.warn(
                "`max_tasks` is ignored when `max_rollouts` is provided.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.max_rollouts = max_rollouts
        self.max_tasks = max_tasks if max_tasks is not None else max_rollouts

        # Note: tracer is created after store to enable auto-detection of OTLP-capable stores
        self._tracer_spec = tracer  # Store spec for later resolution

        if adapter is not None and triplet_exporter is not None:
            warnings.warn(
                "`triplet_exporter` is deprecated and ignored because `adapter` is provided.",
                DeprecationWarning,
                stacklevel=2,
            )

        adapter_spec = adapter if adapter is not None else triplet_exporter
        self.adapter = self._make_adapter(adapter_spec)
        self.triplet_exporter = self.adapter  # Backwards compatibility

        self.algorithm = self._make_algorithm(algorithm)

        # We might be able to support a list of resources in future.
        self.initial_resources = initial_resources

        self.port = port

        self.strategy = self._make_strategy(
            strategy,
            n_runners=self.n_runners,
            port=port,
        )

        # The active store for the current execution context
        self.store = self._make_store(store, self.strategy)

        # Create tracer after store so we can auto-detect OTLP-capable stores
        self.tracer = self._make_tracer(self._tracer_spec, self.store)

        self.runner = self._make_runner(runner)

        if hasattr(self.strategy, "n_runners"):
            strategy_runners = getattr(self.strategy, "n_runners")
            if isinstance(strategy_runners, int) and strategy_runners > 0:
                self.n_runners = strategy_runners
                self.n_workers = strategy_runners

        self.llm_proxy = self._make_llm_proxy(llm_proxy, store=self.store)

        self.hooks = self._normalize_hooks(hooks)

        if not self.daemon:
            logger.warning(
                "daemon=False. Worker processes are non-daemonic. "
                "The worker processes will NOT be terminated when the main process exits. "
                "The cleanup must be handled manually."
            )

    def _make_tracer(
        self, tracer: ComponentSpec[Tracer], store: Optional[LightningStore] = None
    ) -> Tracer:
        """Resolve the tracer component from user input.

        Auto-detects OTLP-capable stores (like InsightLightningStore) and uses OtelTracer
        to enable native OTLP trace export. Falls back to AgentOpsTracer otherwise.
        """
        # Check if store supports OTLP traces (e.g., InsightLightningStore)
        store_supports_otlp = False
        if store is not None:
            # Check store's capabilities
            if hasattr(store, "capabilities"):
                store_supports_otlp = store.capabilities.get("otlp_traces", False)
            # Also check listeners on the store for OTLP capability
            elif hasattr(store, "_listeners"):
                for listener in getattr(store, "_listeners", []):
                    if hasattr(listener, "capabilities"):
                        if listener.capabilities.get("otlp_traces", False):
                            store_supports_otlp = True
                            break

        if store_supports_otlp:
            logger.info("OTLP-capable store detected. Using OtelTracer for native trace export.")
            default_factory = lambda: OtelTracer()
        else:
            default_factory = lambda: AgentOpsTracer(
                agentops_managed=True,
                instrument_managed=True,
                daemon=self.daemon,
            )

        return build_component(
            tracer,
            expected_type=Tracer,
            spec_name="tracer",
            default_factory=default_factory,
            dict_requires_type=True,
            invalid_spec_error_fmt="Invalid tracer type: {actual_type}. Expected Tracer, str, dict, or None.",
            type_error_fmt="Tracer factory returned {type_name}, which is not a Tracer subclass.",
        )

    def _make_algorithm(self, algorithm: ComponentSpec[Algorithm]) -> Optional[Algorithm]:
        """Resolve the algorithm component, allowing `None` for dev-mode dry runs."""
        return build_component(
            algorithm,
            expected_type=Algorithm,
            spec_name="algorithm",
            allow_none=True,
            invalid_spec_error_fmt="Invalid algorithm type: {actual_type}. Expected Algorithm, str, dict, or None.",
            type_error_fmt="Algorithm factory returned {type_name}, which is not a Algorithm subclass.",
        )

    def _make_adapter(self, adapter: ComponentSpec[TraceAdapter[Any]]) -> TraceAdapter[Any]:
        """Resolve the adapter used to transform spans into algorithm-ready payloads."""
        return build_component(
            adapter,
            expected_type=TraceAdapter,
            spec_name="adapter",
            default_factory=TracerTraceToTriplet,
            dict_requires_type=False,
            dict_default_cls=TracerTraceToTriplet,
            invalid_spec_error_fmt="Invalid adapter type: {actual_type}. Expected TraceAdapter, dict, or None.",
            type_error_fmt="Adapter factory returned {type_name}, which is not a TraceAdapter subclass.",
        )

    def _make_store(self, store: ComponentSpec[LightningStore], strategy: ExecutionStrategy) -> LightningStore:
        """Resolve the store implementation backing rollouts, attempts, spans, and resources.

        By default, it's always a in-memory store. If using a client/server execution strategy,
        the in-memory store will be initialized in a thread-safe manner.
        """
        is_client_server = isinstance(strategy, ClientServerExecutionStrategy)
        default_store_factory = lambda: InMemoryLightningStore(thread_safe=is_client_server)
        return build_component(
            store,
            expected_type=LightningStore,
            spec_name="store",
            default_factory=default_store_factory,
            invalid_spec_error_fmt="Invalid store type: {actual_type}. Expected LightningStore, str, dict, or None.",
            type_error_fmt="Store factory returned {type_name}, which is not a LightningStore subclass.",
        )

    def _make_strategy(
        self,
        strategy: ComponentSpec[ExecutionStrategy],
        *,
        n_runners: int,
        port: Optional[int] = None,
    ) -> ExecutionStrategy:
        """Resolve the execution strategy and seed defaults such as `n_runners`."""
        if isinstance(strategy, ExecutionStrategy):
            if port is not None and isinstance(strategy, ClientServerExecutionStrategy):
                strategy.server_port = port
            return strategy
        optional_defaults: Dict[str, Callable[[], Any]] = {"n_runners": lambda: n_runners}
        if port is not None:
            optional_defaults["server_port"] = lambda: port

        def default_factory() -> ExecutionStrategy:
            if port is not None:
                return ClientServerExecutionStrategy(n_runners=n_runners, server_port=port)
            return ClientServerExecutionStrategy(n_runners=n_runners)

        return build_component(
            strategy,
            expected_type=ExecutionStrategy,
            spec_name="strategy",
            default_factory=default_factory,
            optional_defaults=optional_defaults,
            invalid_spec_error_fmt="Invalid strategy type: {actual_type}. Expected ExecutionStrategy, str, dict, or None.",
            type_error_fmt="Strategy factory returned {type_name}, which is not an ExecutionStrategy subclass.",
            registry=ExecutionStrategyRegistry,
        )

    def _make_llm_proxy(
        self,
        llm_proxy: ComponentSpec[LLMProxy],
        *,
        store: LightningStore,
    ) -> Optional[LLMProxy]:
        """Resolve an optional LLM proxy and ensure it shares the trainer's store instance."""
        if isinstance(llm_proxy, LLMProxy):
            return llm_proxy

        optional_defaults: Dict[str, Callable[[], Any]] = {"store": lambda: store}
        
        # Extract OTLP endpoint and headers from store if it supports OTLP
        if hasattr(store, "capabilities") and store.capabilities.get("otlp_traces", False):
            try:
                otlp_endpoint = store.otlp_traces_endpoint()
                otlp_headers = store.get_otlp_headers() if hasattr(store, "get_otlp_headers") else {}
                optional_defaults["otlp_endpoint"] = lambda: otlp_endpoint
                optional_defaults["otlp_headers"] = lambda: otlp_headers
                logger.info(f"Trainer: Store provides OTLP endpoint for proxy: {otlp_endpoint}")
            except Exception as e:
                logger.warning(f"Trainer: Failed to get OTLP endpoint from store: {e}")
        
        if isinstance(llm_proxy, dict):
            llm_proxy = {**llm_proxy}
            llm_proxy.setdefault("store", store)
            # Pass OTLP config to proxy if store provides it and it's not already set
            if "otlp_endpoint" in optional_defaults and "otlp_endpoint" not in llm_proxy:
                llm_proxy.setdefault("otlp_endpoint", optional_defaults["otlp_endpoint"]())
            if "otlp_headers" in optional_defaults and "otlp_headers" not in llm_proxy:
                llm_proxy.setdefault("otlp_headers", optional_defaults["otlp_headers"]())

        return build_component(
            llm_proxy,
            expected_type=LLMProxy,
            spec_name="llm_proxy",
            allow_none=True,
            optional_defaults=optional_defaults,
            invalid_spec_error_fmt="Invalid llm_proxy type: {actual_type}. Expected LLMProxy, dict, str, or None.",
            type_error_fmt="llm_proxy factory returned {type_name}, which is not an LLMProxy subclass.",
        )

    def _make_runner(self, runner: ComponentSpec[Runner[Any]]) -> Runner[Any]:
        """Resolve the runner responsible for executing the agent inside each worker."""
        optional_defaults: Dict[str, Callable[[], Any]] = {"tracer": lambda: self.tracer}
        if self.max_rollouts is not None:
            optional_defaults["max_rollouts"] = lambda: self.max_rollouts

        def default_runner_factory() -> Runner[Any]:
            return instantiate_component(LitAgentRunner, optional_defaults=optional_defaults)

        return build_component(
            runner,
            expected_type=Runner,
            spec_name="runner",
            default_factory=default_runner_factory,
            optional_defaults=optional_defaults,
            invalid_spec_error_fmt="Invalid runner type: {actual_type}. Expected Runner, callable, str, dict, or None.",
            type_error_fmt="Runner factory returned {type_name}, which is not a Runner subclass.",
        )

    def _normalize_hooks(self, hooks: Optional[Union[Hook, Sequence[Hook]]]) -> Sequence[Hook]:
        """Coerce hook inputs into an immutable sequence for runner initialization."""
        if hooks is None:
            return ()
        if isinstance(hooks, Hook):
            return (hooks,)
        return tuple(hooks)

    def fit(
        self,
        agent: LitAgent[T_co],
        train_dataset: Optional[Dataset[T_co]] = None,
        *,
        val_dataset: Optional[Dataset[T_co]] = None,
    ) -> None:
        """Execute the full algorithm/runner training loop.

        [`Trainer.fit`][mantisdk.Trainer.fit] packages the algorithm and runner bundles,
        then hands them to the active [`ExecutionStrategy`][mantisdk.ExecutionStrategy].
        The strategy rarely returns until:

        * The algorithm exhausts the dataset(s) and stops enqueuing rollouts.
        * `max_rollouts` causes individual runners to exit.
        * An exception or interrupt cancels the shared [`ExecutionEvent`][mantisdk.ExecutionEvent].

        Args:
            agent: [`LitAgent`][mantisdk.LitAgent] implementation executed by runners.
            train_dataset: Optional iterable of rollout inputs consumed by the algorithm.
            val_dataset: Optional iterable consumed by validation passes.
        """
        if isinstance(train_dataset, str):
            logger.warning(
                "Trainer.fit will no longer accepts a string URL in future version. "
                "To continue using a string URL, please use Trainer.fit_v0 instead. "
                "See documentation for how to migrate to latest version: https://microsoft.github.io/mantisdk/stable/"
            )
            return self.fit_v0(  # type: ignore
                agent,
                train_dataset,
                val_dataset,  # type: ignore
            )

        agent.set_trainer(self)

        algorithm_bundle = functools.partial(
            self._algorithm_bundle,
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            algorithm=self.algorithm,
        )
        runner_bundle = functools.partial(self._runner_bundle, agent=agent)

        self.strategy.execute(algorithm_bundle, runner_bundle, self.store)

    def dev(
        self,
        agent: LitAgent[T_co],
        train_dataset: Optional[Dataset[T_co]] = None,
        *,
        val_dataset: Optional[Dataset[T_co]] = None,
    ) -> None:
        """Exercise the infrastructure using a fast, synchronous algorithm.

        [`Trainer.dev`][mantisdk.Trainer.dev] mirrors [`fit()`][mantisdk.Trainer.fit] but
        insists on an [`Algorithm`][mantisdk.Algorithm] subtype that also derives from
        [`FastAlgorithm`][mantisdk.FastAlgorithm]. This keeps the loop responsive for
        debugging while still touching the same store, runners, hooks, and tracer plumbing.

        If no algorithm is provided, a default [`Baseline`][mantisdk.Baseline] algorithm will be used.

        Args:
            agent: [`LitAgent`][mantisdk.LitAgent] implementation to execute.
            train_dataset: Optional iterable passed to the algorithm.
            val_dataset: Optional iterable passed to the algorithm.

        Raises:
            TypeError: If the configured algorithm does not inherit from `FastAlgorithm`.
        """
        agent.set_trainer(self)

        # Sanity check
        if self.algorithm is None:
            algorithm = Baseline()
        else:
            algorithm = self.algorithm

        if not isinstance(algorithm, FastAlgorithm):
            raise TypeError(
                "Trainer.dev() requires an algorithm that inherits from FastAlgorithm. "
                f"Received {type(algorithm).__name__}."
            )

        algorithm_bundle = functools.partial(
            self._algorithm_bundle,
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            algorithm=algorithm,
        )
        runner_bundle = functools.partial(self._runner_bundle, agent=agent)
        self.strategy.execute(algorithm_bundle, runner_bundle, self.store)

    async def _algorithm_bundle(
        self,
        store: LightningStore,
        event: ExecutionEvent,
        train_dataset: Optional[Dataset[T_co]],
        val_dataset: Optional[Dataset[T_co]],
        algorithm: Optional[Algorithm],
    ) -> None:
        """Internal entry point executed by the strategy for the algorithm role.

        This coroutine is scheduled inside the strategy's process/thread and is responsible
        for binding algorithm dependencies (store, adapter, initial resources, proxy) before
        invoking [`Algorithm.run`][mantisdk.Algorithm.run].
        When `algorithm` is `None` the bundle simply waits for the
        shared `event` to signal shutdown so runners can still execute (useful for manual queue
        seeding or external algorithms).
        """
        if algorithm is not None:
            algorithm.set_trainer(self)
            algorithm.set_store(store)
            algorithm.set_adapter(self.adapter)
            if self.initial_resources is not None:
                algorithm.set_initial_resources(self.initial_resources)
            if self.llm_proxy is not None:
                self.llm_proxy.set_store(store)
                algorithm.set_llm_proxy(self.llm_proxy)

        if algorithm is None:
            while not event.is_set():
                await asyncio.sleep(0.1)
            return
        try:
            if algorithm.is_async():
                await algorithm.run(  # type: ignore
                    train_dataset=train_dataset,
                    val_dataset=val_dataset,
                )
            else:
                # This will block the event loop to maximize the debugging experience
                # It's the responsibility of the execution strategy to enable async execution
                algorithm.run(
                    train_dataset=train_dataset,
                    val_dataset=val_dataset,
                )
        except Exception:
            logger.exception("Algorithm bundle encountered an error.")
            raise

    async def _runner_bundle(
        self, store: LightningStore, worker_id: int, event: ExecutionEvent, agent: LitAgent[T_co]
    ) -> None:
        """Internal entry point executed by the strategy for each runner role.

        The bundle materializes the configured runner, binds the agent and hooks, associates
        the worker with the shared store, and then drives the runner's [`iter`][mantisdk.Runner.iter]
        loop until the execution event is set or an exception occurs. Cleanup mirrors the initialization
        sequence to keep tracer state, hooks, and agent resources consistent across restarts.
        """
        runner_instance: Runner[Any] | None = None
        runner_initialized = False
        worker_initialized = False
        try:
            # If not using shm execution strategy, we are already in the forked process
            runner_instance = self.runner
            runner_instance.init(agent=agent, hooks=self.hooks)
            runner_initialized = True
            runner_instance.init_worker(worker_id, store)
            worker_initialized = True
            await runner_instance.iter(event=event)
        except Exception:
            logger.exception("Runner bundle encountered an error (worker_id=%s).", worker_id)
            raise
        finally:
            if runner_instance is not None:
                if worker_initialized:
                    try:
                        runner_instance.teardown_worker(worker_id)
                    except Exception:
                        logger.exception("Error during runner worker teardown (worker_id=%s).", worker_id)
                if runner_initialized:
                    try:
                        runner_instance.teardown()
                    except Exception:
                        logger.exception("Error during runner teardown (worker_id=%s).", worker_id)



# Copyright (c) Microsoft. All rights reserved.

from __future__ import annotations

import logging
import time
from contextlib import contextmanager
from typing import (
    Iterator,
    Optional,
)

from mantisdk.types import (
    Attributes,
    SpanCoreFields,
    SpanRecordingContext,
    StatusCode,
    TraceStatus,
)
from mantisdk.utils.otel import format_exception_attributes

from .base import Tracer

logger = logging.getLogger(__name__)


class DummySpanRecordingContext(SpanRecordingContext):
    """Context for recording operations on a dummy span, not dependent on any backend tracer."""

    def __init__(self, name: str, attributes: Optional[Attributes] = None, start_time: Optional[float] = None) -> None:
        self.name = name
        self.attributes = attributes or {}
        self.start_time = start_time or time.time()
        self.end_time = None
        self.status = TraceStatus(status_code="OK")

    def record_exception(self, exception: BaseException) -> None:
        self.record_status("ERROR", str(exception))
        self.record_attributes(format_exception_attributes(exception))

    def record_attributes(self, attributes: Attributes) -> None:
        self.attributes.update(attributes)

    def record_status(self, status_code: StatusCode, description: Optional[str] = None) -> None:
        self.status = TraceStatus(status_code=status_code, description=description)

    def finalize(self, end_time: Optional[float] = None) -> None:
        self.end_time = end_time or time.time()

    def get_recorded_span(self) -> SpanCoreFields:
        if self.end_time is None:
            raise ValueError("End time is not set. Call finalize() first.")
        return SpanCoreFields(
            name=self.name,
            attributes=self.attributes,
            start_time=self.start_time,
            end_time=self.end_time,
            status=self.status,
        )


class DummyTracer(Tracer):
    """A dummy tracer that does not trace anything, but it is compatible with the emitter API.

    It doesn't rely on any backend tracer, and also doesn't use any stores.
    """

    def create_span(
        self,
        name: str,
        attributes: Optional[Attributes] = None,
        timestamp: Optional[float] = None,
        status: Optional[TraceStatus] = None,
    ) -> SpanCoreFields:
        if attributes is None:
            attributes = {}
        if timestamp is None:
            timestamp = time.time()
        if status is None:
            status = TraceStatus(status_code="OK")
        return SpanCoreFields(
            name=name,
            attributes=attributes,
            start_time=timestamp,
            end_time=timestamp,
            status=status,
        )

    @contextmanager
    def operation_context(
        self,
        name: str,
        attributes: Optional[Attributes] = None,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
    ) -> Iterator[DummySpanRecordingContext]:
        start_time = start_time or time.time()
        recording_context = DummySpanRecordingContext(name, attributes, start_time)
        try:
            yield recording_context
        except Exception as exc:
            recording_context.record_exception(exc)
            recording_context.record_status("ERROR", str(exc))
            raise
        finally:
            recording_context.finalize(end_time)
